__all__ = ('InteractionForm',)

from scarletio import RichAttributeErrorBaseType

from ..shared_helpers import create_auto_custom_id

from .fields import (
    parse_components, parse_custom_id, parse_title, put_components, put_custom_id, put_title, validate_components,
    validate_custom_id, validate_title
)


class InteractionForm(RichAttributeErrorBaseType):
    """
    Form component.
    
    Attributes
    ----------
    components : ``None | tuple<Component>``
        Stored components.
    
    custom_id : `None | str`
        Custom identifier to match the form data when receiving it's interaction back.
    
    title : `None | str`
        The form's title.
    """
    __slots__ = ('components', 'custom_id', 'title')
    
    def __new__(cls, title, components, custom_id = None):
        """
        Creates a form interaction.
        
        Parameters
        ----------
        title : `None | str`
            The form's title.
        
        components : ``None | iterable<Component | (tuple | list)<Component>>``
            Sub components.
        
        custom_id : `None | str` = `None`, Optional
             Custom identifier for the form.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        components = validate_components(components)
        custom_id = validate_custom_id(custom_id)
        title = validate_title(title)
        
        if (custom_id is None):
            custom_id = create_auto_custom_id()
        
        self = object.__new__(cls)
        self.components = components
        self.custom_id = custom_id
        self.title = title
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new interaction form from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Interaction form data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.components = parse_components(data)
        self.custom_id = parse_custom_id(data)
        self.title = parse_title(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the interaction form to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_components(self.components, data, defaults)
        put_custom_id(self.custom_id, data, defaults)
        put_title(self.title, data, defaults)
        
        return data
    

    def __repr__(self):
        """Returns the interaction form's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # System fields : custom_id
        
        field_added = False
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            repr_parts.append(' custom_id = ')
            repr_parts.append(repr(custom_id))
        
        # Text fields : label & placeholder
        
        # title
        title = self.title
        if (title is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            repr_parts.append(' title = ')
            repr_parts.append(repr(title))
        
        # sub-component fields : components
        
        # components
        if field_added:
            repr_parts.append(',')
        else:
            field_added = True
        repr_parts.append(' components = ')
        components = self.components
        if (components is None):
            repr_parts.append('[]')
        else:
            repr_parts.append('[')
            
            index = 0
            limit = len(components)
            
            while True:
                component = components[index]
                index += 1
                
                repr_parts.append(repr(component))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        
        repr_parts.append('>')
        return ''.join(repr_parts)


    def copy(self):
        """
        Copies the interaction form.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        
        # components
        components = self.components
        if (components is not None):
            components = tuple(component.copy() for component in self.components)
        new.components = components
        
        # custom_id
        new.custom_id = self.custom_id
        
        # title
        new.title = self.title
        
        return new
    
    
    def copy_with(self, **kwargs):
        """
        Copies the component and modifies the created one with the given parameters.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            Keyword parameters defining which fields should be overwritten.
        
        Other Parameters
        ----------------
        components : ``None | iterable<Component | (tuple | list)<Component>>``, Optional (Keyword only)
            Sub components.
        
        custom_id : `None | str`, Optional (Keyword only)
            Custom identifier to detect which button was clicked by the user.
        
        title : `None | str`, Optional (Keyword only)
            The form's title.
        
        Returns
        -------
        new : ``InteractionForm``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # components
        try:
            components = kwargs.pop('components')
        except KeyError:
            components = self.components
            if (components is not None):
                components = tuple(component.copy() for component in self.components)
        else:
            components = validate_components(components)
        
        # custom_id
        try:
            custom_id = kwargs.pop('custom_id')
        except KeyError:
            custom_id = self.custom_id
        else:
            custom_id = validate_custom_id(custom_id)
        
        # title
        try:
            title = kwargs.pop('title')
        except KeyError:
            title = self.title
        else:
            title = validate_title(title)
        
        
        if custom_id is None:
            custom_id = create_auto_custom_id()
        
        # Construct
        
        new = object.__new__(type(self))
        new.components = components
        new.custom_id = custom_id
        new.title = title
        return new
    
    
    def __eq__(self, other):
        """Returns whether the two interaction forms are equal."""
        if type(other) is not type(self):
            return NotImplemented
        
        # components
        if self.components != other.components:
            return False
        
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        # title
        if self.title != other.title:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the interaction form's hash value."""
        hash_value = 0
        
        # components
        components = self.components
        if (components is not None):
            hash_value ^= len(components) << 12
            for component in components:
                hash_value ^= hash(component)
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(custom_id)
        
        # title
        title = self.title
        if (title is not None):
            hash_value ^= hash(title)
        
        return hash_value
    
    
    def iter_components(self):
        """
        Iterates over the components of the form.
        
        This method is an iterable generator.
        
        Yields
        ------
        component : ``Component``
        """
        components = self.components
        if (components is not None):
            yield from components



__all__ = ('ChannelMetadataPrivateGroup',)

from scarletio import copy_docs
 
from ...bases import ICON_TYPE_NONE, IconSlot, Slotted
from ...http.urls import build_channel_group_icon_url, build_channel_group_icon_url_as
from ...permission.permission import PERMISSION_GROUP, PERMISSION_GROUP_OWNER, PERMISSION_NONE

from .fields import (
    parse_application_id, parse_name, parse_owner_id, put_application_id, put_name, put_owner_id,
    validate_application_id, validate_name, validate_owner_id
)
from .private_base import ChannelMetadataPrivateBase


CHANNEL_METADATA_ICON = IconSlot('icon', 'icon')


class ChannelMetadataPrivateGroup(ChannelMetadataPrivateBase, metaclass = Slotted):
    """
    Channel metadata for private channels.
    
    Attributes
    ----------
    application_id : `int`
        The application's identifier the channel is managed by.
    icon_hash : `int`
        The channel's icon's hash in `uint128`.
    icon_type : ``iconType``
        The channel's icon's type.
    name : `str`
        The channel's display name. Can be empty string if the channel has no name.
    owner_id : `int`
        The group channel's owner's id.
    users : ``list<ClientUserBase>``
        The users in the channel.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('application_id', 'name', 'owner_id')
    
    icon = CHANNEL_METADATA_ICON
    
    
    def __new__(
        cls,
        *,
        application_id = ...,
        icon = ...,
        name = ...,
        owner_id = ...,
        users = ...,
    ):
        """
        Creates a new private group channel metadata from the given parameters.
        
        Parameters
        ----------
        application_id : `int`, ``Application``, Optional (Keyword only)
            The application's identifier the channel is managed by.
        icon : ``None | str | bytes-like | Icon``, Optional (Keyword only)
            The channel's icon.
        name : `str`, Optional (Keyword only)
            The channel's display name. Can be empty string if the channel has no name.
        owner_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The group channel's owner's id.
        users : `iterable` of ``ClientUserBase``, Optional (Keyword only)
            The users in the channel.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # application_id
        if application_id is ...:
            application_id = 0
        else:
            application_id = validate_application_id(application_id)
        
        # icon
        if icon is ...:
            icon = None
        else:
            icon = cls.icon.validate_icon(icon, allow_data = True)
            
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # owner_id
        if owner_id is ...:
            owner_id = 0
        else:
            owner_id = validate_owner_id(owner_id)
    
        # Construct
        self = ChannelMetadataPrivateBase.__new__(
            cls,
            users = users,
        )
        self.application_id = application_id
        self.icon = icon
        self.name = name
        self.owner_id = owner_id
        return self
    
    
    @classmethod
    @copy_docs(ChannelMetadataPrivateBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            application_id = keyword_parameters.pop('application_id', ...),
            icon = keyword_parameters.pop('icon', ...),
            name = keyword_parameters.pop('name', ...),
            owner_id = keyword_parameters.pop('owner_id', ...),
            users = keyword_parameters.pop('users', ...),
        )
    
    
    @copy_docs(ChannelMetadataPrivateBase.__hash__)
    def __hash__(self):
        hash_value = ChannelMetadataPrivateBase.__hash__(self)
        
        # application_id
        hash_value ^= self.application_id
        
        # icon
        hash_value ^= hash(self.icon)
        
        # name
        name = self.name
        if name:
            hash_value ^= hash(name)
        
        # owner_id
        hash_value ^= self.owner_id
        
        return hash_value
    
    
    @classmethod
    @copy_docs(ChannelMetadataPrivateBase.from_data)
    def from_data(cls, data):
        self = super(ChannelMetadataPrivateGroup, cls).from_data(data)
        
        # application_id
        self.application_id = parse_application_id(data)
        
        return self
    
    
    @copy_docs(ChannelMetadataPrivateBase._created)
    def _created(self, channel_entity, client, strong_cache):
        if strong_cache and (client is not None):
            client.group_channels[channel_entity.id] = channel_entity
    
        
    @copy_docs(ChannelMetadataPrivateBase._delete)
    def _delete(self, channel_entity, client):
        if (client is not None):
            try:
                del client.group_channels[channel_entity.id]
            except KeyError:
                pass
    
    
    @copy_docs(ChannelMetadataPrivateBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ChannelMetadataPrivateBase._is_equal_same_type(self, other):
            return False
        
        # application_id
        if self.application_id != other.application_id:
            return False
        
        # icon
        if self.icon_hash != other.icon_hash:
            return False
        
        if self.icon_type is not other.icon_type:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # owner_id
        if self.owner_id != other.owner_id:
            return False
        
        return True
    
    
    @copy_docs(ChannelMetadataPrivateBase._get_processed_name)
    def _get_processed_name(self):
        name = self.name
        if name:
            return name
        
        users = self.users
        if users:
            return ', '.join([user.name for user in users])
        
        return 'Unnamed'
    
    
    @copy_docs(ChannelMetadataPrivateBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataPrivateBase._update_attributes(self, data)
        
        # application_id
        # Ignore internal
        
        # icon
        self._set_icon(data)
        
        # name
        self.name = parse_name(data)
        
        # owner_id
        self.owner_id = parse_owner_id(data)
    
    
    @copy_docs(ChannelMetadataPrivateBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataPrivateBase._difference_update_attributes(self, data)
        
        # application_id
        # Ignore internal
        
        # icon
        self._update_icon(data, old_attributes)
        
        # name
        name = parse_name(data)
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        # owner_id
        owner_id = parse_owner_id(data)
        if self.owner_id != owner_id:
            old_attributes['owner_id'] = self.owner_id
            self.owner_id = owner_id
        
        return old_attributes

    
    @copy_docs(ChannelMetadataPrivateBase._get_permissions_for)
    def _get_permissions_for(self, channel_entity, user):
        if self.owner_id == user.id:
            return PERMISSION_GROUP_OWNER
        elif user in self.users:
            return PERMISSION_GROUP
        else:
            return PERMISSION_NONE
    
    
    @classmethod
    @copy_docs(ChannelMetadataPrivateBase._from_partial_data)
    def _from_partial_data(cls, data):
        self = super(ChannelMetadataPrivateGroup, cls)._from_partial_data(data)
        
        if (data is not None):
            name = data.get('name', None)
            if (name is not None):
                self.name = name
        
        return self
    
    
    @classmethod
    @copy_docs(ChannelMetadataPrivateBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataPrivateGroup, cls)._create_empty()
        
        self.application_id = 0
        self.name = ''
        self.owner_id = 0
        
        self.icon_hash = 0
        self.icon_type = ICON_TYPE_NONE
        
        return self
    
    
    @copy_docs(ChannelMetadataPrivateBase.copy)
    def copy(self):
        new = ChannelMetadataPrivateBase.copy(self)
        new.application_id = self.application_id
        new.name = self.name
        new.owner_id = self.owner_id
        new.icon_hash = self.icon_hash
        new.icon_type = self.icon_type
        return new
    
    
    def copy_with(
        self,
        *,
        application_id = ...,
        icon = ...,
        name = ...,
        owner_id = ...,
        users = ...,
    ):
        """
        Copies the private group channel metadata with the given fields.
        
        Parameters
        ----------
        application_id : `int`, ``Application``, Optional (Keyword only)
            The application's identifier the channel is managed by.
        icon : ``None | str | bytes-like | Icon``, Optional (Keyword only)
            The channel's icon.
        name : `str`, Optional (Keyword only)
            The channel's display name. Can be empty string if the channel has no name.
        owner_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The group channel's owner's id.
        users : `iterable` of ``ClientUserBase``, Optional (Keyword only)
            The users in the channel.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # application_id
        if application_id is ...:
            application_id = self.application_id
        else:
            application_id = validate_application_id(application_id)
        
        # icon
        if icon is ...:
            icon = self.icon
        else:
            icon = type(self).icon.validate_icon(icon, allow_data = True)
            
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # owner_id
        if owner_id is ...:
            owner_id = self.owner_id
        else:
            owner_id = validate_owner_id(owner_id)
    
        # Construct
        new = ChannelMetadataPrivateBase.copy_with(
            self,
            users = users,
        )
        new.application_id = application_id
        new.icon = icon
        new.name = name
        new.owner_id = owner_id
        return new
    
    
    @copy_docs(ChannelMetadataPrivateBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            application_id = keyword_parameters.pop('application_id', ...),
            icon = keyword_parameters.pop('icon', ...),
            name = keyword_parameters.pop('name', ...),
            owner_id = keyword_parameters.pop('owner_id', ...),
            users = keyword_parameters.pop('users', ...),
        )
    
    
    @copy_docs(ChannelMetadataPrivateBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataPrivateBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        type(self).icon.put_into(self.icon, data, defaults, as_data = not include_internals)
        put_name(self.name, data, defaults)
        
        if include_internals:
            put_application_id(self.application_id, data, defaults)
            put_owner_id(self.owner_id, data, defaults)
        
        return data
    
    
    @copy_docs(ChannelMetadataPrivateBase._get_icon_url)
    def _get_icon_url(self, channel_entity):
        return build_channel_group_icon_url(channel_entity.id, self.icon_type, self.icon_hash)
    
    
    @copy_docs(ChannelMetadataPrivateBase._get_icon_url_as)
    def _get_icon_url_as(self, channel_entity, ext, size):
        return build_channel_group_icon_url_as(channel_entity.id, self.icon_type, self.icon_hash, ext, size)



__all__ = ()

from ...emoji import Emoji, create_partial_emoji_data, create_partial_emoji_from_data
from ...field_parsers import entity_id_parser_factory
from ...field_putters import entity_id_putter_factory
from ...field_validators import (
    entity_id_validator_factory, nullable_entity_validator_factory, nullable_string_validator_factory
)

from .constants import TEXT_LENGTH_MAX


# emoji

def parse_emoji(data):
    """
    Parses out the `emoji` field from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Channel data.
    
    Returns
    -------
    emoji : ``None | Emoji``
    """
    nested_data = data.get('poll_media', None)
    if nested_data is None:
        return None
    
    emoji_data = nested_data.get('emoji', None)
    if emoji_data is None:
        return None
    
    return create_partial_emoji_from_data(emoji_data)


def put_emoji(emoji, data, defaults):
    """
    Puts the `emoji`'s data into the given `data` json serializable object.
    
    Parameters
    ----------
    emoji : ``None | Emoji``
        The emoji to serialize.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (emoji is not None):
        nested_data = data.get('poll_media', None)
        if nested_data is None:
            nested_data = {}
            data['poll_media'] = nested_data
        
        if emoji is None:
            emoji_data = None
        else:
            emoji_data = create_partial_emoji_data(emoji)
        
        nested_data['emoji'] = emoji_data
    
    return data


validate_emoji = nullable_entity_validator_factory('emoji', Emoji)

# id

parse_id = entity_id_parser_factory('answer_id')
put_id = entity_id_putter_factory('answer_id')
validate_id = entity_id_validator_factory('answer_id')

# text

def parse_text(data):
    """
    Parses out a poll answer's text from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Poll answer data.
    
    Returns
    -------
    text : `None | str`
    """
    nested_data = data.get('poll_media', None)
    if nested_data is None:
        return None
    
    text = nested_data.get('text', None)
    if (text is None) or (not text):
        return None
    
    return text


def put_text(text, data, defaults):
    """
    Serializes the given text into the given data.
    
    Parameters
    ----------
    text : `None | str`
        Poll answer text.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    nested_data = data.get('poll_media', None)
    if nested_data is None:
        nested_data = {}
        data['poll_media'] = nested_data
    
    if text is None:
        text = ''
    
    nested_data['text'] = text
    return data


validate_text = nullable_string_validator_factory('text', 0, TEXT_LENGTH_MAX)



__all__ = ()

from scarletio import Compound, Theory

from ...http import DiscordApiClient
from ...user import ClientUserBase, User
from ...utils import log_time_converter

from ..request_helpers import (
    get_channel_id_and_message_id, get_message_and_channel_id_and_message_id, get_poll_answer_and_id
)


POLL_RESULT_USER_GET_CHUNK_LIMIT_MIN = 1
POLL_RESULT_USER_GET_CHUNK_LIMIT_MAX = 100


class ClientCompoundPollEndpoints(Compound):
    
    api : DiscordApiClient
    
    @Theory
    async def message_get(self, message, *, force_update = False): ...
    
    
    async def poll_result_user_get_chunk(self, message, poll_answer, *, after = ..., limit = ...):
        """
        Requests the users, who voted on the given message with the given answer.
        
        If the message has no poll or has no answers for that answer returns empty list. If we know every voters for
        the given answers it will pull them from the cache.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``(int, int) | Message``
            The message whats voters should be requested.
        
        poll_answer : ``PollAnswer``, `int`
            The poll answer to request the users for.
        
        after : ``None | int | DiscordEntity | DateTime``, Optional (Keyword only)
            The timestamp after the voters were created.
        
        limit : `int`, Optional (Keyword only)
            The amount of users to request. Can be in range [1:100].
        
        Returns
        -------
        users : ``list<ClientUserBase>``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        `before` parameter is not supported by Discord.
        """
        if limit is ...:
            limit = POLL_RESULT_USER_GET_CHUNK_LIMIT_MAX
        
        elif isinstance(limit, int):
            if limit < POLL_RESULT_USER_GET_CHUNK_LIMIT_MIN:
                limit = POLL_RESULT_USER_GET_CHUNK_LIMIT_MIN
            elif limit > POLL_RESULT_USER_GET_CHUNK_LIMIT_MAX:
                limit = POLL_RESULT_USER_GET_CHUNK_LIMIT_MAX
        
        else:
            raise TypeError(
                f'`limit` can be `None | int`, got {type(limit).__name__}; {limit!r}.'
            )
        
        message, channel_id, message_id = get_message_and_channel_id_and_message_id(message)
        poll_answer, poll_answer_id = get_poll_answer_and_id(poll_answer)
        if not poll_answer_id:
            if message is None:
                message = await self.message_get((channel_id, message_id),)
            
            poll = message.poll
            if poll is None:
                return []
            
            poll_answer_id = poll.get_answer_id(poll_answer)
            if not poll_answer_id:
                return []
        
        if (message is not None):
            poll = message.poll
            if poll is None:
                return []
            
            result = poll[poll_answer_id]
            if result is None:
                return []
            
            if not result.unknown:
                after = 0 if after is ... else log_time_converter(after)
                # before = 9223372036854775807 if before is ... else log_time_converter(before)
                return result.filter_after(limit, after)
        
        
        query_parameters = {
            'limit': limit,
        }
        
        if (after is not ...):
            query_parameters['after'] = log_time_converter(after)
        
        # Before is not supported!
        # if (before is not ...):
        #     data['before'] = log_time_converter(before)
        
        data = await self.api.poll_result_user_get_chunk(
            channel_id, message_id, poll_answer_id, query_parameters
        )
        
        user_datas = data.get('users', None)
        if user_datas is None:
            users = []
        else:
            users = [User.from_data(user_data) for user_data in user_datas]
        
        if (message is not None):
            poll = message.poll
            if (poll is not None):
                poll._fill_some_votes(poll_answer_id, users)
        
        return users
    
    
    async def poll_result_user_get_all(self, message, poll_answer):
        """
        Requests the all the voters for the given answer.
        
        If all the voters are known for the given answer, they are queried from the cache instead.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``(int, int) | Message``
            The message whats voters should be requested.
        poll_answer : ``PollAnswer``, `int`
            The poll answer to request the users for.
        
        Returns
        -------
        users : ``list<ClientUserBase>``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        message, channel_id, message_id = get_message_and_channel_id_and_message_id(message)
        poll_answer, poll_answer_id = get_poll_answer_and_id(poll_answer)
        
        if not poll_answer_id:
            if message is None:
                message = await self.message_get((channel_id, message_id),)
            
            poll = message.poll
            if poll is None:
                return []
            
            poll_answer_id = poll.get_answer_id(poll_answer)
            if not poll_answer_id:
                return []
        
        if (message is not None):
            poll = message.poll
            if poll is None:
                return []
            
            result = poll[poll_answer_id]
            if result is None:
                return []
            
            if not result.unknown:
                return sorted(result)
        
        query_parameters = {
            'after': 0,
            'limit': POLL_RESULT_USER_GET_CHUNK_LIMIT_MAX,
        }
        
        users = []
        
        while True:
            data = await self.api.poll_result_user_get_chunk(
                channel_id, message_id, poll_answer_id, query_parameters
            )
            user_datas = data.get('users', None)
            if (user_datas is not None):
                for user_data in user_datas:
                    users.append(User.from_data(user_data))
            
            if (user_datas is None) or (len(user_datas) < POLL_RESULT_USER_GET_CHUNK_LIMIT_MAX):
                break
            
            query_parameters['after'] = users[-1].id
        
        if (message is not None):
            poll = message.poll
            if (poll is not None):
                poll._fill_all_votes(poll_answer_id, users)
        
        return users
    
    
    async def poll_result_get_all(self, message):
        """
        Requests all the voters for every answer on the given message.
        
        Like the other poll voter getting methods, this method prefers using the internal cache as well over doing a
        request.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``(int, int) | Message``
            The message, what's voters should be requested.
        
        Returns
        -------
        message : ``Message``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        # Get message
        message = await self.message_get(message)
        
        # Get poll
        poll = message.poll
        if (poll is not None):
            users = []
            query_parameters = {
                'limit': POLL_RESULT_USER_GET_CHUNK_LIMIT_MAX,
            }
            
            for answer, result in poll.iter_items():
                if not result.unknown:
                    continue
                
                query_parameters['after'] = 0
                
                while True:
                    data = await self.api.poll_result_user_get_chunk(
                        message.channel_id, message.id, answer.id, query_parameters
                    )
                    
                    user_datas = data.get('users', None)
                    if user_datas is not None:                    
                        for user_data in user_datas:
                            users.append(User.from_data(user_data))
                    
                    if (user_datas is None) or (len(user_datas) < POLL_RESULT_USER_GET_CHUNK_LIMIT_MAX):
                        break
                    
                    query_parameters['after'] = users[-1].id
                
                result._fill_all_votes(users)
                users.clear()
        
        return message
    
    
    async def poll_finalize(self, message):
        """
        Finalizes the poll on the given message.

        This method is a coroutine.
        
        Parameters
        ----------
        message : ``(int, int) | Message``
            The message, to finalize the poll on it.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id, message_id = get_channel_id_and_message_id(message)
        await self.api.poll_finalize(channel_id, message_id)



__all__ = ()

from scarletio import Compound, Theory

from ...core import MESSAGES
from ...emoji import Emoji
from ...emoji.reaction_mapping.fields import validate_reaction
from ...http import DiscordApiClient
from ...user import ClientUserBase, User
from ...utils import log_time_converter

from ..request_helpers import get_channel_id_and_message_id, get_reaction_emoji_value_and_type, get_user_id


class ClientCompoundReactionEndpoints(Compound):
    
    id: int
    api : DiscordApiClient
    
    @Theory
    async def message_get(self, message, *, force_update = False): ...
    
    
    async def reaction_add(self, message, reaction):
        """
        Adds a reaction on the given message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, `tuple` (`int`, `int`)
            The message on which the reaction will be put on.
        
        emoji : ``Reaction``, ``Emoji``, `str`
            The reaction to react with. if given as ``Emoji`` / `str` will use standard reaction.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id, message_id = get_channel_id_and_message_id(message)
        emoji_value, reaction_type = get_reaction_emoji_value_and_type(reaction)
        await self.api.reaction_add(channel_id, message_id, emoji_value, {'type': reaction_type.value})
    
    
    async def reaction_delete(self, message, reaction, user):
        """
        Removes the specified reaction of the user from the given message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, `tuple` (`int`, `int`)
            The message from which the reaction will be removed.
            
        reaction : ``Reaction``, ``Emoji``, `str`
            The reaction to remove.
            
        user : ```ClientUserBase``, `int`
            The user, who's reaction will be removed.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id, message_id = get_channel_id_and_message_id(message)
        emoji_value, reaction_type = get_reaction_emoji_value_and_type(reaction)
        user_id = get_user_id(user)
        
        if user_id == self.id:
            await self.api.reaction_delete_own(channel_id, message_id, emoji_value, reaction_type.value)
        else:
            await self.api.reaction_delete(channel_id, message_id, emoji_value, reaction_type.value, user_id)
    
    
    async def reaction_delete_emoji(self, message, emoji):
        """
        Removes all the reaction of the specified emoji from the given message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, `tuple` (`int`, `int`)
            The message from which the reactions will be removed.
            
        emoji : ``Emoji``, `str`
            The reaction to remove.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id, message_id = get_channel_id_and_message_id(message)
        
        if isinstance(emoji, Emoji):
            emoji_value = emoji.as_reaction
        elif isinstance(emoji, str):
            emoji_value = emoji
        else:
            raise TypeError(
                f'`emoji` can be `{Emoji.__name__}`, `str`, got {emoji.__class__.__name__}; {emoji!r}.'
            )
        
        await self.api.reaction_delete_emoji(channel_id, message_id, emoji_value)
    
    
    async def reaction_delete_own(self, message, reaction):
        """
        Removes the specified reaction of the client from the given message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, `tuple` (`int`, `int`)
            The message from which the reaction will be removed.
            
        emoji : ``Reaction``, ``Emoji``, `str`
            The reaction to remove.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id, message_id = get_channel_id_and_message_id(message)
        emoji_value, reaction_type = get_reaction_emoji_value_and_type(reaction)
        await self.api.reaction_delete_own(channel_id, message_id, emoji_value, reaction_type.value)
    
    
    async def reaction_clear(self, message):
        """
        Removes all the reactions from the given message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, `tuple` (`int`, `int`)
            The message from which the reactions will be cleared.
        
        Raises
        ------
        TypeError
            If `message` was not given neither as ``Message``, `tuple` (`int`, `int`) instance.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id, message_id = get_channel_id_and_message_id(message)
        
        await self.api.reaction_clear(channel_id, message_id)
    
    
    async def reaction_user_get_chunk(self, message, reaction, *, limit = None, after = None):
        """
        Requests the users, who reacted on the given message with the given emoji.
        
        If the message has no reactors at all or no reactors with that emoji, returns an empty list. If we know the
        emoji's every reactors we query the parameters from that.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, `tuple` (`int`, `int`)
            The message, what's reactions will be requested.
        reaction : ``Reaction``, ``Emoji``, `str`
            The emoji, what's reactors will be requested.
        limit : `None | int` = `None`, Optional (Keyword only)
            The amount of users to request. Can be in range [1:100]. Defaults to 100 by Discord.
        after : ``None | int | DiscordEntity | DateTime`` = `None`, Optional (Keyword only)
            The timestamp after the message's reactors were created.
        
        Returns
        -------
        users : ``list<ClientUserBase>``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            The given `emoji` is not a valid reaction.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        `before` parameter is not supported by Discord.
        """
        if limit is None:
            limit = 100
        
        elif isinstance(limit, int):
            if limit < 1:
                limit = 1
            elif limit > 100:
                limit = 100
        
        else:
            raise TypeError(
                f'`limit` can be `None | int`, got {type(limit).__name__}; {limit!r}.'
            )
        
        channel_id, message_id = get_channel_id_and_message_id(message)
        reaction = validate_reaction(reaction)
        
        try:
            message = MESSAGES[message_id]
        except KeyError:
            reactions = None
        else:
            reactions = message.reactions
            
            if reactions is None:
                return []
            
            try:
                line = reactions[reaction]
            except KeyError:
                return []
            
            if not line.unknown:
                after = 0 if after is None else log_time_converter(after)
                # before = 9223372036854775807 if before is None else log_time_converter(before)
                users = line.filter_after(limit, after)
                return users
        
        query_parameters = {
            'limit': limit,
            'type': reaction.type.value,
        }
        
        if (after is not None):
            query_parameters['after'] = log_time_converter(after)
        
        # Before is not supported!
        # if (before is not None):
        #     data['before'] = log_time_converter(before)
        
        data = await self.api.reaction_user_get_chunk(
            channel_id, message_id, reaction.emoji.as_reaction, query_parameters
        )
        
        users = [User.from_data(user_data) for user_data in data]
        
        if (reactions is not None):
            reactions._fill_some_reactions(reaction, users)
        
        return users
    
    
    async def reaction_user_get_all(self, message, reaction):
        """
        Requests the all the users, which reacted on the message with the given message.
        
        If the message has no reactors at all or no reactors with that emoji returns an empty list. If the emoji's
        every reactors are known, then do requests are done.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, `tuple` (`int`, `int`)
            The message, what's reactions will be requested.
        reaction : ``Reacton``, ``Emoji``, `str`
            The reaction, what's reactors will be requested.
        
        Returns
        -------
        users : ``list<ClientUserBase>``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            The given `emoji` is not a valid reaction.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id, message_id = get_channel_id_and_message_id(message)
        reaction = validate_reaction(reaction)
        
        try:
            message = MESSAGES[message_id]
        except KeyError:
            reactions = None
        else:
            reactions = message.reactions
            
            if (reactions is None):
                return []
            
            try:
                line = reactions[reaction]
            except KeyError:
                return []
            
            if not line.unknown:
                users = [*line]
                return users
        
        query_parameters = {
            'after': 0,
            'limit': 100,
            'type': reaction.type.value,
        }
        
        users = []
        emoji_value = reaction.emoji.as_reaction
        
        while True:
            user_datas = await self.api.reaction_user_get_chunk(
                channel_id, message_id, emoji_value, query_parameters
            )
            users.extend(User.from_data(user_data) for user_data in user_datas)
            
            if len(user_datas) < 100:
                break
            
            query_parameters['after'] = users[-1].id
        
        if (reactions is not None):
            reactions._fill_all_reactions(reaction, users)
        
        return users
    
    
    async def reaction_get_all(self, message):
        """
        Requests all the reactors for every emoji on the given message.
        
        Like the other reaction getting methods, this method prefers using the internal cache as well over doing a
        request.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, `tuple` (`int`, `int`)
            The message, what's reactions will be requested.
        
        Returns
        -------
        message : ``Message``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        # Get message
        message = await self.message_get(message)
        
        # Get reactions
        reactions = message.reactions
        if (reactions is not None) and reactions:
            users = []
            query_parameters = {
                'limit': 100,
            }
            
            for reaction, line in reactions.iter_items():
                if not line.unknown:
                    continue
                
                emoji_value = reaction.emoji.as_reaction
                query_parameters['after'] = 0
                query_parameters['type'] = reaction.type.value
                
                while True:
                    user_datas = await self.api.reaction_user_get_chunk(
                        message.channel_id, message.id, emoji_value, query_parameters
                    )
                    users.extend(User.from_data(user_data) for user_data in user_datas)
                    
                    if len(user_datas) < 100:
                        break
                    
                    query_parameters['after'] = users[-1].id
                
                reactions._fill_all_reactions(reaction, users)
                users.clear()
        
        return message



"""
Hata extensions come into this directory.

If a hata library extension is imported, it should register itself with the ``register_library_extension``, so if an
other library extension is imported, it can register a hook to run only if both are imported.

Hook registrations can be done with the ``add_library_extension_hook`` function.
"""
__all__ = ()

from warnings import warn
from importlib.util import find_spec
from importlib import import_module


def __getattr__(attribute_name):
    """Gets the extension from name if not yet imported."""
    
    spec = find_spec(f'{__name__}.{attribute_name}')
    if (spec is not None):
        try:
            module = import_module(spec.name)
        except ImportError:
            raise
        
        except BaseException as err:
            raise ImportError from err
        
        else:
            return module
    
    raise AttributeError(attribute_name)


from scarletio.ext import (
    add_library_extension_hook, register_library_extension
)


EXTENSION_SETUP_FUNCTIONS = {}


class SetupFunction:
    """
    Describes a setup function.
    
    Attributes
    ----------
    extension_name : `str`
        The represented extension's full name.
    extension_short_name : `str`
        The represented extension's short name.
    optional_parameters : `None | tuple<str>`
        Optional parameters of the `setup_function`.
    required_parameters : `None | tuple<str>`
        Required parameters by the `setup_function`.
    setup_function : ``FunctionType``
        The setup function itself.
    """
    __slots__ = (
        'extension_name', 'extension_short_name', 'optional_parameters', 'required_parameters', 'setup_function'
    )
    
    def __new__(cls, extension_name, setup_function, required_parameters, optional_parameters):
        """
        Creates a new ``SetupFunction`` from the given parameters.
        
        Parameters
        ----------
        extension_name : `str`
            The represented extension's full name.
        setup_function : ``FunctionType``
            The setup function itself.
        required_parameters : `None | tuple<str>`
            Required parameters by the `setup_function`.
        optional_parameters : `None | tuple<str>`
            Optional parameters of the `setup_function`.
        """
        dot_index = extension_name.find('.')
        if dot_index != -1:
            extension_short_name = extension_name[dot_index + 1:]
        else:
            extension_short_name = extension_name
        
        self = object.__new__(cls)
        self.setup_function = setup_function
        self.required_parameters = required_parameters
        self.optional_parameters = optional_parameters
        self.extension_name = extension_name
        self.extension_short_name = extension_short_name
        return self
    
    
    def __call__(self, client, input_keyword_parameters):
        """
        Calls the internal ``setup_function``.
        
        Parameters
        ----------
        client : ``Client``
            The client on who the extension should be setupped.
        input_keyword_parameters : `dict<str, object>`
            Keyword parameters to get the extension's parameters from.
        """
        positional_parameters = []
        required_parameters = self.required_parameters
        if (required_parameters is not None):
            for required_parameter in required_parameters:
                positional_parameters.append(input_keyword_parameters[required_parameter])
        
        keyword_parameters = {}
        optional_parameters = self.optional_parameters
        if (optional_parameters is not None):
            for optional_parameter in optional_parameters:
                try:
                    parameter_value = input_keyword_parameters[optional_parameter]
                except KeyError:
                    pass
                else:
                    keyword_parameters[optional_parameter] = parameter_value
        
        self.setup_function(client, *positional_parameters, **keyword_parameters)


def register_setup_function(extension_name, setup_function, required_parameters, optional_parameters):
    """
    Registers an extension setup function.
    
    Parameters
    ----------
    extension_name : `str`
        The extension's system name.
    setup_function : `FunctionType``
        The setup function of the extension.
    required_parameters : `None | tuple<str>`
        Required parameters by the `setup_function`.
    optional_parameters : `None | tuple<str>`
        Optional parameters of the `setup_function`.
    """
    setup_function = SetupFunction(extension_name, setup_function, required_parameters, optional_parameters)
    
    EXTENSION_SETUP_FUNCTIONS[extension_name] = setup_function
    EXTENSION_SETUP_FUNCTIONS.setdefault(setup_function.extension_short_name, setup_function)


def _try_get_setup_function(extension_name, extension_short_name):
    """
    Tries to get setup function for the given extension name.
    
    Parameters
    ----------
    extension_name : `str`
        The extension's name.
    
    Returns
    -------
    setup_function : `None`, ``SetupFunction``
        A function to setup the respective extension on a client.
    """
    try:
        setup_function = EXTENSION_SETUP_FUNCTIONS[extension_name]
    except KeyError:
        if (extension_short_name is None):
            setup_function = None
        else:
            setup_function = EXTENSION_SETUP_FUNCTIONS.get(extension_short_name)
    
    return setup_function


def _get_setup_function(extension_name):
    """
    Gets extension setup function for the given extension name.
    
    Parameters
    ----------
    extension_name : `str`
        The extension's name.
    
    Returns
    -------
    setup_function : ``SetupFunction``
        A function to setup the respective extension on a client.
    
    Raises
    ------
    ImportError
        if importing the extension failed.
    ModuleNotFoundError
        If the extension not found.
    RuntimeError
        If the extension has no setup function.
    """
    dot_index = extension_name.find('.')
    if dot_index == -1:
        short_name = None
    else:
        short_name = extension_name[dot_index + 1:]
    
    setup_function = _try_get_setup_function(extension_name, short_name)
    if (setup_function is not None):
        return setup_function
    
    # Use goto
    while True:
        spec = find_spec(f'{__name__}.{extension_name}')
        if (spec is not None):
            break
        
        if (short_name is not None):
            spec = find_spec(f'{__name__}.{short_name}')
            if (spec is not None):
                break
        
        raise ModuleNotFoundError(extension_name)
    
    try:
        import_module(spec.name)
    except ImportError:
        raise
    except BaseException as err:
        raise ImportError from err
    
    setup_function = _try_get_setup_function(extension_name, short_name)
    if (setup_function is None):
        raise RuntimeError(f'`Extension: {extension_name!r} has no setup function.')
    
    return setup_function


def get_setup_functions(extensions, keyword_parameters):
    """
    Gets and validates setup function parameters.
    
    Parameters
    ----------
    extensions : `None | set<str>`
        The extension(s)'s name to setup on a client.
    keyword_parameters : `dict<str, object>`
        Additional parameters to pass to extensions.
    
    Returns
    -------
    setup_functions : `None`, `set` of ``SetupFunction``
        Setup functions to setup on a client.
    
    Raises
    ------
    ImportError
        If importing an extension failed.
    ModuleNotFoundError
        If an extension not found.
    RuntimeError
        - If `keyword_parameters` not contains any required parameter.
        - The an extension has no setup function.
    
    Notes
    -----
    If `keyword_parameters` contains any extra parameters, `RuntimeWarning` is dropped.
    
    > The function may block if new extension is imported, so please consider avoiding production time calls.
    """
    setup_functions = None
    
    if (extensions is not None):
        for extension_name in extensions:
            setup_function = _get_setup_function(extension_name)
            
            if setup_functions is None:
                setup_functions = set()
            
            setup_functions.add(setup_function)
    
    exhaustible_parameters = set(keyword_parameters.keys())
    
    if (setup_functions is not None):
        for setup_function in setup_functions:
            required_parameters = setup_function.required_parameters
            if (required_parameters is not None):
                exhaustible_parameters.difference_update(required_parameters)
                
                for required_parameter in required_parameters:
                    if required_parameter not in keyword_parameters:
                        raise RuntimeError(
                            f'`{required_parameter}` parameter is required by '
                            f'`{setup_function.extension_short_name}`.'
                        )
            
            optional_parameters = setup_function.optional_parameters
            if (optional_parameters is not None):
                exhaustible_parameters.difference_update(optional_parameters)
    
    if exhaustible_parameters:
        warn(
            (
                f'`get_setup_functions` received unused parameters: '
                f'{", ".join(f"{name} = {keyword_parameters[name]!r}" for name in exhaustible_parameters)}.'
            ),
            RuntimeWarning,
            stacklevel = 3,
        )
    
    return setup_functions


def run_setup_functions(client, setup_functions, keyword_parameters):
    """
    Setups the given extensions on the client.
    
    Parameters
    ----------
    client : ``Client``
        The client on who the extensions should be setupped.
    setup_functions : `None`, `set` of ``SetupFunction``
        The setup functions to run with the client.
    keyword_parameters : `dict<str, object>`
        Keyword parameters to get the extensions's parameters from.
    """
    if (setup_functions is not None):
        for setup_function in setup_functions:
            setup_function(client, keyword_parameters)



__all__ = ('ApplicationCommandOptionMetadataPrimitive',)

from scarletio import copy_docs

from .fields import (
    parse_autocomplete, parse_choices, put_autocomplete, put_choices, validate_autocomplete,
    validate_choices_postprocessed
)
from .parameter import ApplicationCommandOptionMetadataParameter


class ApplicationCommandOptionMetadataPrimitive(ApplicationCommandOptionMetadataParameter):
    """
    Base type for primitive parameter application command option metadatas.
    
    Attributes
    ----------
    autocomplete : `bool`
        Whether the option supports auto completion.
    
    choices : `None`, `tuple` of ``ApplicationCommandOptionChoice``
        Choices for the user to pick from.
    
    required : `bool`
        Whether the parameter is required. Defaults to `False`.
    """
    TYPE = object
    
    __slots__ = ('autocomplete', 'choices')
    
    
    def __new__(cls, *, autocomplete = ..., choices = ..., required = ...):
        """
        Creates a new primitive application command option metadata with the given parameters.
        
        Parameters
        ----------
        autocomplete : `bool`, Optional (Keyword only)
            Whether the option supports auto completion.
        
        choices : `None`, `iterable` of ``ApplicationCommandOptionChoice``, Optional (Keyword only)
            Choices for the user to pick from.
        
        required : `bool`, Optional (Keyword only)
            Whether the parameter is required.
        
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        ValueError
            - If a parameter of incorrect value given.
        """
        # autocomplete
        if autocomplete is ...:
            autocomplete = False
        else:
            autocomplete = validate_autocomplete(autocomplete)
        
        # choices
        if choices is ...:
            choices = None
        else:
            choices = validate_choices_postprocessed(choices, cls.TYPE)
        
        # Construct
        new = ApplicationCommandOptionMetadataParameter.__new__(
            cls,
            required = required,
        )
        new.autocomplete = autocomplete
        new.choices = choices
        return new
    
    
    @classmethod
    @copy_docs(ApplicationCommandOptionMetadataParameter.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            autocomplete = keyword_parameters.pop('autocomplete', ...),
            choices = keyword_parameters.pop('choices', ...),
            required = keyword_parameters.pop('required', ...),
        )
    
    @classmethod
    @copy_docs(ApplicationCommandOptionMetadataParameter.from_data)
    def from_data(cls, data):
        self = super(ApplicationCommandOptionMetadataPrimitive, cls).from_data(data)
        self.autocomplete = parse_autocomplete(data)
        self.choices = parse_choices(data)
        return self
    
    
    @copy_docs(ApplicationCommandOptionMetadataParameter.to_data)
    def to_data(self, *, defaults = False):
        data = ApplicationCommandOptionMetadataParameter.to_data(self, defaults = defaults)
        put_autocomplete(self.autocomplete, data, defaults)
        put_choices(self.choices, data, defaults)
        return data
    
    
    @copy_docs(ApplicationCommandOptionMetadataParameter._add_type_specific_repr_fields)
    def _add_type_specific_repr_fields(self, repr_parts):
        ApplicationCommandOptionMetadataParameter._add_type_specific_repr_fields(self, repr_parts)
        
        repr_parts.append(', autocomplete = ')
        repr_parts.append(repr(self.autocomplete))
        
        repr_parts.append(', choices = ')
        repr_parts.append(repr(self.choices))
    
    
    @copy_docs(ApplicationCommandOptionMetadataParameter._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ApplicationCommandOptionMetadataParameter._is_equal_same_type(self, other):
            return False
        
        if self.autocomplete != other.autocomplete:
            return False
        
        if self.choices != other.choices:
            return False
        
        return True
    
    
    @copy_docs(ApplicationCommandOptionMetadataParameter.__hash__)
    def __hash__(self):
        hash_value = ApplicationCommandOptionMetadataParameter.__hash__(self)
        
        # autocomplete
        hash_value ^= self.autocomplete << 1
        
        # choices
        choices = self.choices
        if (choices is not None):
            hash_value ^= len(choices) << 1
            
            for choice in choices:
                hash_value ^= hash(choice)
        
        return hash_value
    
    
    @copy_docs(ApplicationCommandOptionMetadataParameter.copy)
    def copy(self):
        new = ApplicationCommandOptionMetadataParameter.copy(self)
        new.autocomplete = self.autocomplete
        
        # choices
        choices = self.choices
        if (choices is not None):
            choices = (*(choice.copy() for choice in choices),)
        new.choices = choices
        
        return new
    
    
    def copy_with(self, *, autocomplete = ..., choices = ..., required = ...):
        """
        Copies the primitive application command option metadata with the given fields.
        
        Parameters
        ----------
        autocomplete : `bool`, Optional (Keyword only)
            Whether the option supports auto completion.
        
        choices : `None`, `iterable` of ``ApplicationCommandOptionChoice``, Optional (Keyword only)
            Choices for the user to pick from.
        
        required : `bool`, Optional (Keyword only)
            Whether the parameter is required.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        ValueError
            - If a parameter of incorrect value given.
        """
        # autocomplete
        if autocomplete is ...:
            autocomplete = self.autocomplete
        else:
            autocomplete = validate_autocomplete(autocomplete)
        
        # choices
        if choices is ...:
            choices = self.choices
            if (choices is not None):
                choices = (*(choice.copy() for choice in choices),)
        else:
            choices = validate_choices_postprocessed(choices, type(self).TYPE)
        
        # Construct
        new = ApplicationCommandOptionMetadataParameter.copy_with(
            self,
            required = required,
        )
        new.autocomplete = autocomplete
        new.choices = choices
        return new
    
    
    @copy_docs(ApplicationCommandOptionMetadataParameter.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            autocomplete = keyword_parameters.pop('autocomplete', ...),
            choices = keyword_parameters.pop('choices', ...),
            required = keyword_parameters.pop('required', ...),
        )



__all__ = ()

from scarletio import RichAttributeErrorBaseType


class ArrayUInt32BE(RichAttributeErrorBaseType):
    """
    Implements an uint32 array casted on bytes.
    Using `big endian` byte order.
    
    Attributes
    ----------
    _data : `bytes`
        The source `bytes` object.
    
    _start : `int`
        The first byte what is inside of the array.
    
    _end : `int`
        The first byte, what is not inside of the array after `._start`.
    """
    __slots__ = ('_data', '_end', '_start')
    
    def __new__(cls, data, start, end):
        """
        Creates a new uint32 array from the given parameters.
        
        Parameters
        ----------
        data : `bytes`
            The source `bytes` object.
        start : `int`
            The first byte what is inside of the array.
        end : `int`
            The first byte, what is not inside of the array after `._start`.
        """
        self = object.__new__(cls)
        self._data = data
        self._end = end
        self._start = start
        return self
    
    
    def __len__(self):
        """Returns the array's length"""
        return (self._end - self._start) >> 2
    
    
    def __getitem__(self, index):
        """Returns the element of the array at the given index."""
        location = self._start + (index << 2)
        return int.from_bytes(self._data[location : location + 4], 'big')
    
    
    def __repr__(self):
        """Returns the array's representation."""
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append(' length = ')
        repr_parts.append(repr(len(self)))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two arrays are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return memoryview(self._data)[self._start : self._end] == memoryview(other._data)[other._start : other._end]
    
    
    def __iter__(self):
        """Iterated over the array's elements."""
        data = self._data
        location = self._start
        end = self._end
        
        while True:
            if location >= end:
                break
            
            value = int.from_bytes(data[location : location + 4], 'big')
            yield value
            
            location += 4
            continue



from .base import *
from .constants import *
from .fields import *
from .flags import *
from .guild_announcements import *
from .guild_base import *
from .guild_category import *
from .guild_directory import *
from .guild_forum import *
from .guild_forum_base import *
from .guild_main_base import *
from .guild_media import *
from .guild_stage import *
from .guild_store import *
from .guild_text import *
from .guild_text_base import *
from .guild_thread_announcements import *
from .guild_thread_base import *
from .guild_thread_private import *
from .guild_thread_public import *
from .guild_voice import *
from .guild_voice_base import *
from .preinstanced import *
from .private import *
from .private_base import *
from .private_group import *


__all__ = (
    *base.__all__,
    *constants.__all__,
    *fields.__all__,
    *flags.__all__,
    *guild_announcements.__all__,
    *guild_base.__all__,
    *guild_category.__all__,
    *guild_directory.__all__,
    *guild_forum.__all__,
    *guild_forum_base.__all__,
    *guild_main_base.__all__,
    *guild_media.__all__,
    *guild_stage.__all__,
    *guild_store.__all__,
    *guild_text.__all__,
    *guild_text_base.__all__,
    *guild_thread_announcements.__all__,
    *guild_thread_base.__all__,
    *guild_thread_private.__all__,
    *guild_thread_public.__all__,
    *guild_voice.__all__,
    *guild_voice_base.__all__,
    *preinstanced.__all__,
    *private.__all__,
    *private_base.__all__,
    *private_group.__all__,
)



__all__ = ()

from ...emoji.emoji.fields import validate_available, validate_name, validate_role_ids

from ..audit_log_entry_change_conversion import AuditLogEntryChangeConversion, AuditLogEntryChangeConversionGroup
from ..conversion_helpers.converters import (
    value_deserializer_ids, value_deserializer_name, value_serializer_ids, value_serializer_name
)


# ---- available ----

AVAILABLE_CONVERSION = AuditLogEntryChangeConversion(
    ('available',),
    'available',
    value_validator = validate_available,
)


@AVAILABLE_CONVERSION.set_value_deserializer
def available_value_deserializer(value):
    if value is None:
        value = True
    return value


# ---- name ----

NAME_CONVERSION = AuditLogEntryChangeConversion(
    ('name',),
    'name',
    value_deserializer = value_deserializer_name,
    value_serializer = value_serializer_name,
    value_validator = validate_name,
)


# ---- role_ids ----

ROLE_IDS_CONVERSION = AuditLogEntryChangeConversion(
    ('roles',),
    'role_ids',
    value_deserializer = value_deserializer_ids,
    value_serializer = value_serializer_ids,
    value_validator = validate_role_ids,
)


# ---- Construct ----

EMOJI_CONVERSIONS = AuditLogEntryChangeConversionGroup(
    AVAILABLE_CONVERSION,
    NAME_CONVERSION,
    ROLE_IDS_CONVERSION,
)



__all__ = ('ForumTagUpdate',)

from itertools import count

from scarletio import RichAttributeErrorBaseType

from ..forum_tag import ForumTag

from .fields import validate_forum_tag, validate_old_attributes


class ForumTagUpdate(RichAttributeErrorBaseType):
    """
    Represents an updated forum tag with storing the forum tag and it's old updated attributes in a `dict`.
    
    Attributes
    ----------
    forum_tag : ``ForumTag``
        The updated forum tag.
    old_attributes : `dict<str, object>`
        The changed attributes of the forum tag in `attribute-name` - `old-value` relation. Can contain any of the
        following items:
        
        +-----------+-------------------+
        | Keys      | Values            |
        +===========+===================+
        | emoji     | ``None | Emoji``  |
        +-----------+-------------------+
        | name      | `str`             |
        +-----------+-------------------+
        | moderated | `bool`            |
        +-----------+-------------------+
    """
    __slots__ = ('forum_tag', 'old_attributes',)
    
    def __new__(cls, forum_tag = ..., old_attributes = ...):
        """
        Creates a new forum tag change instance with the given fields.
        
        Parameters
        ----------
        forum_tag : ``ForumTag``, Optional (Keyword only)
            The updated forum tag.
        old_attributes : `None`, `dict<str, object>`, Optional (Keyword only)
            The changed attributes of the forum tag.
        
        Raises
        ------
        TypeError
            - If a field's type is incorrect.
        """
        # forum_tag
        if forum_tag is ...:
            forum_tag = ForumTag()
        else:
            forum_tag = validate_forum_tag(forum_tag)
        
        # old_attributes
        if old_attributes is ...:
            old_attributes = {}
        else:
            old_attributes = validate_old_attributes(old_attributes)
        
        # Construct
        self = object.__new__(cls)
        self.forum_tag = forum_tag
        self.old_attributes = old_attributes
        return self
    
    
    @classmethod
    def from_fields(cls, forum_tag, old_attributes):
        """
        Creates a new forum tag change instance with the given fields.
        
        Parameters
        ----------
        forum_tag : ``ForumTag``
            The updated forum tag.
        old_attributes : `dict<str, object>`
            The changed attributes of the forum tag.
        
        Returns
        -------
        new : `instance<cls>`
        """
        self = object.__new__(cls)
        self.forum_tag = forum_tag
        self.old_attributes = old_attributes
        return self
    
    
    def __repr__(self):
        """Returns the representation of the forum tag update."""
        return f'<{self.__class__.__name__} changes: {len(self.old_attributes)!r}, forum_tag = {self.forum_tag!r}>'
    
    
    def __hash__(self):
        """Returns the forum tag update's hash value."""
        hash_value = 0
        
        # forum_tag
        hash_value ^= hash(self.forum_tag)
        
        # old_attributes
        old_attributes = self.old_attributes
        hash_value ^= len(old_attributes)
        
        for mask, key in zip(count(5, 7), sorted(old_attributes.keys())):
            hash_value ^= (mask | hash(key)) & hash(old_attributes[key])
        
        return hash_value
    
    
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.forum_tag != other.forum_tag:
            return False
        
        if self.old_attributes != other.old_attributes:
            return False
        
        return True
    
    
    def __len__(self):
        """Helper for unpacking if needed."""
        return 2
    
    
    def __iter__(self):
        """
        Unpacks the forum tag update.
        
        This method is a generator.
        """
        yield self.forum_tag
        yield self.old_attributes
    
    
    def copy(self):
        """
        Copies the forum_tag update.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.forum_tag = self.forum_tag.copy()
        new.old_attributes = self.old_attributes.copy()
        return new
    
    
    def copy_with(self, *, forum_tag = ..., old_attributes = ...):
        """
        Copies the forum tag change with the given fields.
        
        Parameters
        ----------
        forum_tag : ``ForumTag``, Optional (Keyword only)
            The updated forum_tag.
        old_attributes : `None`, `dict<str, object>`, Optional (Keyword only)
            The changed attributes of the forum tag.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a field's type is incorrect.
        """
        # forum_tag
        if forum_tag is ...:
            forum_tag = self.forum_tag.copy()
        else:
            forum_tag = validate_forum_tag(forum_tag)
        
        # old_attributes
        if old_attributes is ...:
            old_attributes = self.old_attributes.copy()
        else:
            old_attributes = validate_old_attributes(old_attributes)
        
        # Construct
        new = object.__new__(type(self))
        new.forum_tag = forum_tag
        new.old_attributes = old_attributes
        return new



from .channel import *
from .channel_metadata import *
from .forum_tag import *
from .forum_tag_change import *
from .forum_tag_update import *
from .permission_overwrite import *
from .voice_channel_effect import *

from .message_history import *
from .message_iterator import *


__all__ = (
    *channel.__all__,
    *channel_metadata.__all__,
    *forum_tag.__all__,
    *forum_tag_change.__all__,
    *forum_tag_update.__all__,
    *permission_overwrite.__all__,
    *voice_channel_effect.__all__,
    
    *message_history.__all__,
    *message_iterator.__all__,
)



__all__ = ()

from os import get_terminal_size
from math import floor, log10

from ...core import REGISTERED_COMMANDS, REGISTERED_COMMANDS_BY_NAME, command_sort_key, normalize_command_name, register
from ...core.helpers import render_main_call_into


BREAK_LINE_LENGTH_DEFAULT = 80


def render_break_line_into(into, character, name):
    """
    Renders a break line with the given name at the middle.
    
    Parameters
    ----------
    into : `list` of `str`
        The container to render into.
    character : `str`
        The character to create the break from.
    name : `str`
        The command's name.
    
    Returns
    -------
    into : `list` of `str`
    """
    try:
        terminal_size = get_terminal_size()
    except OSError:
        break_line_length = BREAK_LINE_LENGTH_DEFAULT
    else:
        break_line_length = terminal_size.columns
    
    leftover_space = break_line_length - len(name) - 2
    if leftover_space <= 0:
        into.append(name)
    else:
        into.append(character * (leftover_space >> 1))
        into.append(' ')
        into.append(name)
        into.append(' ')
        into.append(character * ((leftover_space >> 1) + (leftover_space & 1)))
    
    into.append('\n')
    
    return into


def show_command(all_, command_name, sub_command_names):
    """
    Returns a command's description.
    
    Parameters
    ----------
    all : `bool`
        Whether all commands should be scanned for the specified name.
    command_name : `str`
        The command name to query for.
    sub_command_names : `tuple` of `str`
        Additional sub command names to search for within the main command.
    
    Returns
    -------
    description : `str`
    """
    command = REGISTERED_COMMANDS_BY_NAME.get(normalize_command_name(command_name), None)
    if (not all_) and (not command.available):
        command = None
    
    output_parts = []
    if command is None:
        output_parts = []
        
        output_parts.append('No command is added for: ')
        output_parts.append(repr(command_name))
        output_parts.append('\n\n')
        output_parts.append('Try using "$ ')
        render_main_call_into(output_parts)
        output_parts.append(' help" to list ')
        output_parts.append('all the' if all_ else 'the available')
        output_parts.append(' commands.\n')
    
    else:
        command.render_direct_usage_into(output_parts, *sub_command_names)
    
    return ''.join(output_parts)


def list_commands(all_):
    """
    Lists all commands.
    
    Parameters
    ----------
    all : `bool`
        Whether all commands should be shown.
    
    Returns
    -------
    description : `str`
    """
    output_parts = []
    output_parts.append('All' if all_ else 'Available')
    output_parts.append(' commands:\n\n')
    
    command_count = len(REGISTERED_COMMANDS)
    if command_count:
        index_adjust = floor(log10(command_count)) + 1
    else:
        index_adjust = 1
    
    if all_:
        commands = [*REGISTERED_COMMANDS]
    else:
        commands = [command for command in REGISTERED_COMMANDS if command.available]
    commands.sort(key = command_sort_key)
    
    for index, command in enumerate(commands, 1):
        command_name = command.name
        output_parts.append(str(index).rjust(index_adjust))
        output_parts.append('. ')
        output_parts.append(command_name)
        output_parts.append('\n')
    
    output_parts.append('\n')
    output_parts.append('Use "$ ')
    render_main_call_into(output_parts)
    output_parts.append(' help COMMAND-NAME')
    if all_:
        output_parts.append(' --all')
    output_parts.append('" for more information.\n')
    
    return ''.join(output_parts)


@register(
    name = 'help',
    aliases = 'h',
)
def help_(command_name: str = None, *sub_command_names, all_: bool = False):
    """Either lists the available commands, or shows the command's usage."""
    if command_name is None:
        return list_commands(all_)
    else:
        return show_command(all_, command_name, sub_command_names)


@register(
    into = help_,
)
def list_(*, all_: bool = False):
    """
    Shows the help of the available commands with their sub-commands included.
    Use the `--all` parameter to show the non-available ones as well.
    """
    output_parts = []
    
    if all_:
        commands = [*REGISTERED_COMMANDS]
    else:
        commands = [command for command in REGISTERED_COMMANDS if command.available]
    commands.sort(key = command_sort_key)
    
    for command in commands:
        command_name = command.name
        output_parts = render_break_line_into(output_parts, '=', command_name)
        
        for command_function in command.iter_command_functions(sort = True):
            command_function_name = command_function.get_full_name()
            if command_function_name != command_name:
                output_parts = render_break_line_into(output_parts, '-', command_function_name)
            
            output_parts.append('\n')
            output_parts = command_function.render_usage_into(output_parts)
            output_parts.append('\n')
    
    return ''.join(output_parts)



__all__ = ()

from scarletio import RichAttributeErrorBaseType

from .descriptor import _conversion_descriptor_sort_key


class SerializationConfiguration(RichAttributeErrorBaseType):
    """
    Configuration representing which fields of a builder and how should be serialised.
    
    Attributes
    ----------
    conversions : `tuple<Conversion>`
        Conversions.
    defaults : `bool`
        Whether defaults should be included.
    """
    __slots__ = ('conversions', 'defaults')
    
    def __new__(cls, conversion_descriptors, defaults):
        """
        Creates a new serialisation configuration.
        
        Parameters
        ----------
        conversion_descriptors : `list<ConversionDescriptor>`
            Conversion descriptors.
        defaults : `bool`
            Whether defaults should be included.
        """
        conversion_descriptors = sorted(conversion_descriptors, key = _conversion_descriptor_sort_key)
        conversions = (*(descriptor.conversion for descriptor in conversion_descriptors),)
        
        self = object.__new__(cls)
        self.conversions = conversions
        self.defaults = defaults
        return self
    
    
    def __repr__(self):
        """Returns the serialization configuration's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # conversions
        repr_parts.append(' conversions = ')
        repr_parts.append(repr(self.conversions))
        
        # defaults
        repr_parts.append(' defaults = ')
        repr_parts.append(repr(self.defaults))
        
        repr_parts.append('>')
        return ''.join(repr_parts)



__all__ = ('ChannelMetadataGuildForumBase',)

from scarletio import copy_docs

from ...permission import Permission
from ...permission.permission import PERMISSION_MASK_VIEW_CHANNEL, PERMISSION_NONE, PERMISSION_THREAD_AND_VOICE_DENY

from ..forum_tag import ForumTag
from ..forum_tag_change import ForumTagChange
from ..forum_tag_update import ForumTagUpdate

from .constants import AUTO_ARCHIVE_DEFAULT, SLOWMODE_DEFAULT
from .fields import (
    parse_available_tags, parse_default_forum_layout, parse_default_sort_order, parse_default_thread_auto_archive_after,
    parse_default_thread_reaction_emoji, parse_default_thread_slowmode, parse_flags, parse_topic, put_available_tags,
    put_default_forum_layout, put_default_sort_order, put_default_thread_auto_archive_after,
    put_default_thread_reaction_emoji, put_default_thread_slowmode, put_flags, put_topic,
    validate_available_tags, validate_default_forum_layout, validate_default_sort_order,
    validate_default_thread_auto_archive_after, validate_default_thread_reaction_emoji, validate_default_thread_slowmode,
    validate_flags, validate_topic
)
from .flags import ChannelFlag
from .guild_main_base import ChannelMetadataGuildMainBase
from .preinstanced import ForumLayout, SortOrder


class ChannelMetadataGuildForumBase(ChannelMetadataGuildMainBase):
    """
    Base type for guild forum-like channel metadatas.
    
    Attributes
    ----------
    _cache_permission : ``None | dict<int, Permission>``
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    available_tags : `None`, `tuple` of ``ForumTag``
        The available tags to assign to the child-thread channels.
    default_forum_layout : ``ForumLayout``
        The default layout used to display threads of the forum.
    default_sort_order : ``SortOrder``
        The default thread ordering of the forum.
    default_thread_auto_archive_after : `int`
        The default duration (in seconds) for newly created threads to automatically archive the themselves. Defaults
        to `3600`. Can be one of: `3600`, `86400`, `259200`, `604800`.
    default_thread_reaction_emoji : ``None | Emoji``
        The emoji to show in the add reaction button on a thread of the forum channel.
    default_thread_slowmode : `int`
        The default slowmode applied to the channel's threads.
    flags : ``ChannelFlag``
        The channel's flags.
    name : `str`
        The channel's name.
    parent_id : `int`
        The channel's parent's identifier.
    permission_overwrites : ``None | dict<int, PermissionOverwrite>``
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    topic : `None`, `str`
        The channel's topic.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = (
        'available_tags', 'default_forum_layout', 'default_sort_order', 'default_thread_auto_archive_after',
        'default_thread_reaction_emoji', 'default_thread_slowmode', 'flags', 'topic',
    )
    
    def __new__(
        cls,
        *,
        available_tags = ...,
        default_forum_layout = ...,
        default_sort_order = ...,
        default_thread_auto_archive_after = ...,
        default_thread_reaction_emoji = ...,
        default_thread_slowmode = ...,
        flags = ...,
        name = ...,
        parent_id = ...,
        permission_overwrites = ...,
        position = ...,
        topic = ...,
    ):
        """
        Creates a new guild forum channel metadata from the given parameters.
        
        Parameters
        ----------
        available_tags : `None`, `iterable` of ``ForumTag``, Optional (Keyword only)
            The available tags to assign to the child-thread channels.
        default_forum_layout : ``ForumLayout``, `int`, Optional (Keyword only)
            The default layout used to display threads of the forum.
        default_sort_order : ``SortOrder``, `int`, Optional (Keyword only)
            The default thread ordering of the forum.
        default_thread_auto_archive_after : `int`, Optional (Keyword only)
            The default duration (in seconds) for newly created threads to automatically archive the themselves.
        default_thread_reaction_emoji : ``None | Emoji``, Optional (Keyword only)
            The emoji to show in the add reaction button on a thread of the forum channel.
        default_thread_slowmode : `int`, Optional (Keyword only)
            The default slowmode applied to the channel's threads.
        flags : ``ChannelFlag``, `int`, Optional (Keyword only)
            The channel's flags.
        name : `str`, Optional (Keyword only)
            The channel's name.
        parent_id : ``None | int | Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        permission_overwrites : ``None | iterable<PermissionOverwrite>``, Optional (Keyword only)
            The channel's permission overwrites.
        position : `int`, Optional (Keyword only)
            The channel's position.
        topic : `None`, `str`, Optional (Keyword only)
            The channel's topic.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # available_tags
        if available_tags is ...:
            available_tags = None
        else:
            available_tags = validate_available_tags(available_tags)
        
        # default_forum_layout
        if default_forum_layout is ...:
            default_forum_layout = ForumLayout.none
        else:
            default_forum_layout = validate_default_forum_layout(default_forum_layout)
        
        # default_sort_order
        if default_sort_order is ...:
            default_sort_order = SortOrder.latest_activity
        else:
            default_sort_order = validate_default_sort_order(default_sort_order)
        
        # default_thread_auto_archive_after
        if default_thread_auto_archive_after is ...:
            default_thread_auto_archive_after = AUTO_ARCHIVE_DEFAULT
        else:
            default_thread_auto_archive_after = validate_default_thread_auto_archive_after(
                default_thread_auto_archive_after
            )
        
        # default_thread_reaction_emoji
        if default_thread_reaction_emoji is ...:
            default_thread_reaction_emoji = None
        else:
            default_thread_reaction_emoji = validate_default_thread_reaction_emoji(default_thread_reaction_emoji)
        
        # default_thread_slowmode
        if default_thread_slowmode is ...:
            default_thread_slowmode = SLOWMODE_DEFAULT
        else:
            default_thread_slowmode = validate_default_thread_slowmode(default_thread_slowmode)
        
        # flags
        if flags is ...:
            flags = ChannelFlag()
        else:
            flags = validate_flags(flags)
        
        # topic
        if topic is ...:
            topic = None
        else:
            topic = validate_topic(topic)
        
        # Construct
        self = ChannelMetadataGuildMainBase.__new__(
            cls,
            name = name,
            permission_overwrites = permission_overwrites,
            parent_id = parent_id,
            position = position,
        )
        self.available_tags = available_tags
        self.default_forum_layout = default_forum_layout
        self.default_sort_order = default_sort_order
        self.default_thread_auto_archive_after = default_thread_auto_archive_after
        self.default_thread_reaction_emoji = default_thread_reaction_emoji
        self.default_thread_slowmode = default_thread_slowmode
        self.flags = flags
        self.topic = topic
        return self
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildMainBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            available_tags = keyword_parameters.pop('available_tags', ...),
            default_forum_layout = keyword_parameters.pop('default_forum_layout', ...),
            default_sort_order = keyword_parameters.pop('default_sort_order', ...),
            default_thread_auto_archive_after = keyword_parameters.pop('default_thread_auto_archive_after', ...),
            default_thread_reaction_emoji = keyword_parameters.pop('default_thread_reaction_emoji', ...),
            default_thread_slowmode = keyword_parameters.pop('default_thread_slowmode', ...),
            flags = keyword_parameters.pop('flags', ...),
            name = keyword_parameters.pop('name', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
            permission_overwrites = keyword_parameters.pop('permission_overwrites', ...),
            position = keyword_parameters.pop('position', ...),
            topic = keyword_parameters.pop('topic', ...),
        )
    
    
    @copy_docs(ChannelMetadataGuildMainBase.__hash__)
    def __hash__(self):
        hash_value = ChannelMetadataGuildMainBase.__hash__(self)
        
        # available_tags
        available_tags = self.available_tags
        if (available_tags is not None):
            hash_value ^= len(available_tags) << 4
            
            for available_tag in available_tags:
                hash_value ^= hash(available_tag)
        
        # default_forum_layout
        hash_value ^= self.default_forum_layout.value << 24
        
        # default_sort_order
        hash_value ^= self.default_sort_order.value << 20
        
        # default_thread_auto_archive_after
        hash_value ^= self.default_thread_auto_archive_after << 16
        
        # default_thread_reaction_emoji
        default_thread_reaction_emoji = self.default_thread_reaction_emoji
        if (default_thread_reaction_emoji is not None):
            hash_value ^= hash(default_thread_reaction_emoji)
        
        # default_thread_slowmode
        hash_value ^= self.default_thread_slowmode << 8
        
        # flags
        hash_value ^= self.flags
        
        # topic
        topic = self.topic
        if (topic is not None):
            hash_value ^= hash(topic)
        
        return hash_value
    
    
    @copy_docs(ChannelMetadataGuildMainBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ChannelMetadataGuildMainBase._is_equal_same_type(self, other):
            return False
        
        # available_tags
        if self.available_tags != other.available_tags:
            return False
        
        # default_forum_layout
        if self.default_forum_layout is not other.default_forum_layout:
            return False
        
        # default_sort_order
        if self.default_sort_order is not other.default_sort_order:
            return False
        
        # default_thread_auto_archive_after
        if self.default_thread_auto_archive_after != other.default_thread_auto_archive_after:
            return False
        
        # default_thread_reaction_emoji
        if self.default_thread_reaction_emoji != other.default_thread_reaction_emoji:
            return False
        
        # default_thread_slowmode
        if self.default_thread_slowmode != other.default_thread_slowmode:
            return False
        
        # flags
        if self.flags != other.flags:
            return False
        
        # topic
        if self.topic != other.topic:
            return False
        
        return True
    
    
    @copy_docs(ChannelMetadataGuildMainBase._get_display_name)
    def _get_display_name(self):
        return self.name.lower()
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildMainBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildForumBase, cls)._create_empty()
        
        self.available_tags = None
        self.default_forum_layout = ForumLayout.none
        self.default_sort_order = SortOrder.latest_activity
        self.default_thread_auto_archive_after = AUTO_ARCHIVE_DEFAULT
        self.default_thread_reaction_emoji = None
        self.default_thread_slowmode = SLOWMODE_DEFAULT
        self.flags = ChannelFlag()
        self.topic = None
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildMainBase.copy)
    def copy(self):
        new = ChannelMetadataGuildMainBase.copy(self)
        
        available_tags = self.available_tags
        if (available_tags is not None):
            available_tags = (*available_tags,)
        new.available_tags = available_tags
        
        new.default_forum_layout = self.default_forum_layout
        new.default_sort_order = self.default_sort_order
        new.default_thread_auto_archive_after = self.default_thread_auto_archive_after
        new.default_thread_reaction_emoji = self.default_thread_reaction_emoji
        new.default_thread_slowmode = self.default_thread_slowmode
        new.flags = self.flags
        new.topic = self.topic
        
        return new
    
    
    def copy_with(
        self,
        *,
        available_tags = ...,
        default_forum_layout = ...,
        default_sort_order = ...,
        default_thread_auto_archive_after = ...,
        default_thread_reaction_emoji = ...,
        default_thread_slowmode = ...,
        flags = ...,
        name = ...,
        parent_id = ...,
        permission_overwrites = ...,
        position = ...,
        topic = ...,
    ):
        """
        Copies the guild forum channel metadata with the given fields.
        
        Parameters
        ----------
        available_tags : `None`, `iterable` of ``ForumTag``, Optional (Keyword only)
            The available tags to assign to the child-thread channels.
        default_forum_layout : ``ForumLayout``, `int`, Optional (Keyword only)
            The default layout used to display threads of the forum.
        default_sort_order : ``SortOrder``, `int`, Optional (Keyword only)
            The default thread ordering of the forum.
        default_thread_auto_archive_after : `int`, Optional (Keyword only)
            The default duration (in seconds) for newly created threads to automatically archive the themselves.
        default_thread_reaction_emoji : ``None | Emoji``, Optional (Keyword only)
            The emoji to show in the add reaction button on a thread of the forum channel.
        default_thread_slowmode : `int`, Optional (Keyword only)
            The default slowmode applied to the channel's threads.
        flags : ``ChannelFlag``, `int`, Optional (Keyword only)
            The channel's flags.
        name : `str`, Optional (Keyword only)
            The channel's name.
        parent_id : ``None | int | Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        permission_overwrites : ``None | iterable<PermissionOverwrite>``, Optional (Keyword only)
            The channel's permission overwrites.
        position : `int`, Optional (Keyword only)
            The channel's position.
        topic : `None`, `str`, Optional (Keyword only)
            The channel's topic.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # available_tags
        if available_tags is ...:
            available_tags = self.available_tags
            if (available_tags is not None):
                available_tags = (*available_tags,)
        else:
            available_tags = validate_available_tags(available_tags)
        
        # default_forum_layout
        if default_forum_layout is ...:
            default_forum_layout = self.default_forum_layout
        else:
            default_forum_layout = validate_default_forum_layout(default_forum_layout)
        
        # default_sort_order
        if default_sort_order is ...:
            default_sort_order = self.default_sort_order
        else:
            default_sort_order = validate_default_sort_order(default_sort_order)
        
        # default_thread_auto_archive_after
        if default_thread_auto_archive_after is ...:
            default_thread_auto_archive_after = self.default_thread_auto_archive_after
        else:
            default_thread_auto_archive_after = validate_default_thread_auto_archive_after(
                default_thread_auto_archive_after
            )
        
        # default_thread_reaction_emoji
        if default_thread_reaction_emoji is ...:
            default_thread_reaction_emoji = self.default_thread_reaction_emoji
        else:
            default_thread_reaction_emoji = validate_default_thread_reaction_emoji(default_thread_reaction_emoji)
        
        # default_thread_slowmode
        if default_thread_slowmode is ...:
            default_thread_slowmode = self.default_thread_slowmode
        else:
            default_thread_slowmode = validate_default_thread_slowmode(default_thread_slowmode)
        
        # flags
        if flags is ...:
            flags = self.flags
        else:
            flags = validate_flags(flags)
        
        # topic
        if topic is ...:
            topic = self.topic
        else:
            topic = validate_topic(topic)
        
        # Construct
        new = ChannelMetadataGuildMainBase.copy_with(
            self,
            name = name,
            permission_overwrites = permission_overwrites,
            parent_id = parent_id,
            position = position,
        )
        new.available_tags = available_tags
        new.default_forum_layout = default_forum_layout
        new.default_sort_order = default_sort_order
        new.default_thread_auto_archive_after = default_thread_auto_archive_after
        new.default_thread_reaction_emoji = default_thread_reaction_emoji
        new.default_thread_slowmode = default_thread_slowmode
        new.flags = flags
        new.topic = topic
        return new
    
    
    @copy_docs(ChannelMetadataGuildMainBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            available_tags = keyword_parameters.pop('available_tags', ...),
            default_forum_layout = keyword_parameters.pop('default_forum_layout', ...),
            default_sort_order = keyword_parameters.pop('default_sort_order', ...),
            default_thread_auto_archive_after = keyword_parameters.pop('default_thread_auto_archive_after', ...),
            default_thread_reaction_emoji = keyword_parameters.pop('default_thread_reaction_emoji', ...),
            default_thread_slowmode = keyword_parameters.pop('default_thread_slowmode', ...),
            flags = keyword_parameters.pop('flags', ...),
            name = keyword_parameters.pop('name', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
            permission_overwrites = keyword_parameters.pop('permission_overwrites', ...),
            position = keyword_parameters.pop('position', ...),
            topic = keyword_parameters.pop('topic', ...),
        )
    
    
    @copy_docs(ChannelMetadataGuildMainBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildMainBase._update_attributes(self, data)
        
        # available_tags
        self.available_tags = parse_available_tags(data)
        
        # default_forum_layout
        self.default_forum_layout = parse_default_forum_layout(data)
        
        # default_sort_order
        self.default_sort_order = parse_default_sort_order(data)
        
        # default_thread_auto_archive_after
        self.default_thread_auto_archive_after = parse_default_thread_auto_archive_after(data)
        
        # default_thread_reaction_emoji
        self.default_thread_reaction_emoji = parse_default_thread_reaction_emoji(data)
        
        # default_thread_slowmode
        self.default_thread_slowmode = parse_default_thread_slowmode(data)
        
        # flags
        self.flags = parse_flags(data)
        
        # topic
        self.topic = parse_topic(data)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildMainBase._difference_update_attributes(self, data)
        
        # available_tags
        forum_tag_change = self._difference_update_available_tags(data)
        if (forum_tag_change is not None):
            old_attributes['available_tags'] = forum_tag_change
        
        # default_forum_layout
        default_forum_layout = parse_default_forum_layout(data)
        if self.default_forum_layout is not default_forum_layout:
            old_attributes['default_forum_layout'] = self.default_forum_layout
            self.default_forum_layout = default_forum_layout
        
        # default_sort_order
        default_sort_order = parse_default_sort_order(data)
        if self.default_sort_order is not default_sort_order:
            old_attributes['default_sort_order'] = self.default_sort_order
            self.default_sort_order = default_sort_order
        
        # default_thread_auto_archive_after
        default_thread_auto_archive_after = parse_default_thread_auto_archive_after(data)
        if self.default_thread_auto_archive_after != default_thread_auto_archive_after:
            old_attributes['default_thread_auto_archive_after'] = self.default_thread_auto_archive_after
            self.default_thread_auto_archive_after = default_thread_auto_archive_after
        
        # default_thread_reaction_emoji
        default_thread_reaction_emoji = parse_default_thread_reaction_emoji(data)
        if self.default_thread_reaction_emoji != default_thread_reaction_emoji:
            old_attributes['default_thread_reaction_emoji'] = self.default_thread_reaction_emoji
            self.default_thread_reaction_emoji = default_thread_reaction_emoji
        
        # default_thread_slowmode
        default_thread_slowmode = parse_default_thread_slowmode(data)
        if self.default_thread_slowmode != default_thread_slowmode:
            old_attributes['default_thread_slowmode'] = self.default_thread_slowmode
            self.default_thread_slowmode = default_thread_slowmode
        
        # flags
        flags = parse_flags(data)
        if (self.flags != flags):
            old_attributes['flags'] = self.flags
            self.flags = flags
        
        # topic
        topic = parse_topic(data)
        if self.topic != topic:
            old_attributes['topic'] = self.topic
            self.topic = topic
        
        return old_attributes
    
    
    def _difference_update_available_tags(self, data):
        """
        Updates the guild forum's available tags and returns a ``ForumTagChange`` instance if anything was modified.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Channel data.
        
        Returns
        -------
        forum_tag_change : `None`, ``ForumTagChange``
        """
        new_available_tag_datas = data.get('available_tags', None)
        if (new_available_tag_datas is not None) and (not new_available_tag_datas):
            new_available_tag_datas = None
        
        old_available_tags = self.available_tags
        if (old_available_tags is None) and (new_available_tag_datas is None):
            return None
        
        
        old_forum_tags = set()
        if (old_available_tags is not None):
            old_forum_tags.update(old_available_tags)
        
        new_forum_tags = set()
        
        # Update
        updated_forum_tags = None
        
        if (new_available_tag_datas is not None):
            for forum_tag_data in new_available_tag_datas:
                forum_tag, old_attributes = ForumTag._create_or_difference_update(forum_tag_data)
                if (old_attributes is not None) and old_attributes:
                    if (updated_forum_tags is None):
                        updated_forum_tags = []
                    
                    updated_forum_tags.append(ForumTagUpdate.from_fields(forum_tag, old_attributes))
                
                new_forum_tags.add(forum_tag)
                continue
        
        # Added
        added_forum_tags = old_forum_tags - new_forum_tags
        if added_forum_tags:
            added_forum_tags = sorted(added_forum_tags)
        else:
            added_forum_tags = None
        
        # Removed
        removed_forum_tags = new_forum_tags - old_forum_tags
        if removed_forum_tags:
            removed_forum_tags = sorted(removed_forum_tags)
        else:
            removed_forum_tags = None
        
        # Set new
        if new_forum_tags:
            new_available_tags = tuple(sorted(new_forum_tags))
        else:
            new_available_tags = None
        self.available_tags = new_available_tags
        
        # Construct
        if (added_forum_tags is None) and (removed_forum_tags is None) and (updated_forum_tags is None):
            return
        
        return ForumTagChange.from_fields(added_forum_tags, updated_forum_tags, removed_forum_tags)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._iter_delete)
    def _iter_delete(self, channel_entity, client):
        for thread in channel_entity.threads:
            thread._delete(client)
            yield thread
        
        yield from ChannelMetadataGuildMainBase._iter_delete(self, channel_entity, client)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._get_permissions_for)
    def _get_permissions_for(self, channel_entity, user):
        result = self._get_base_permissions_for(channel_entity, user)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # forum channels do not have thread and voice related permissions
        result &= PERMISSION_THREAD_AND_VOICE_DENY
        
        return Permission(result)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._get_permissions_for_roles)
    def _get_permissions_for_roles(self, channel_entity, roles):
        result = self._get_base_permissions_for_roles(channel_entity, roles)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # forum channels do not have thread and voice related permissions
        result &= PERMISSION_THREAD_AND_VOICE_DENY
        return Permission(result)
    
    
    @copy_docs(ChannelMetadataGuildMainBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataGuildMainBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        # available_tags
        put_available_tags(self.available_tags, data, defaults, include_internals = include_internals)
        
        # default_forum_layout
        put_default_forum_layout(self.default_forum_layout, data, defaults)
        
        # default_sort_order
        put_default_sort_order(self.default_sort_order, data, defaults)
        
        # default_auto_archive_duration
        put_default_thread_auto_archive_after(self.default_thread_auto_archive_after, data, defaults)
        
        # default_thread_reaction_emoji
        put_default_thread_reaction_emoji(self.default_thread_reaction_emoji, data, defaults)
        
        # default_thread_slowmode
        put_default_thread_slowmode(self.default_thread_slowmode, data, defaults)
        
        # flags
        put_flags(self.flags, data, defaults)
        
        # topic
        put_topic(self.topic, data, defaults)
        
        return data



__all__ = ()

from .preinstanced import AutoModerationRuleTriggerType


KEYWORD_TO_RULE_TRIGGER_TYPE = {
    'excluded_keywords': None, # This is applicable for 2 types, so we ignore it
    'keywords': AutoModerationRuleTriggerType.keyword,
    'regex_patterns': AutoModerationRuleTriggerType.keyword,
    'keyword_presets': AutoModerationRuleTriggerType.keyword_preset,
    'mention_limit': AutoModerationRuleTriggerType.mention_spam,
}


def guess_rule_trigger_type_from_keyword_parameters(rule_trigger_type, keyword_parameters):
    """
    Guesses the rule trigger's type from the given fields.
    
    Parameters
    ----------
    rule_trigger_type : ``AutoModerationRuleTriggerType``
        Already detected rule trigger type.
    
    keyword_parameters : `dict<str, object>`
        Keyword parameters passed.
    
    Returns
    -------
    rule_trigger_type : ``AutoModerationRuleTriggerType``
        The detected type.
    
    Raises
    ------
    TypeError
        - Extra keyword parameter.
        - Multiple trigger types received / detected.
    """
    for key in keyword_parameters.keys():
        try:
            key_rule_trigger_type = KEYWORD_TO_RULE_TRIGGER_TYPE[key]
        except KeyError:
            raise TypeError(
                f'Extra keyword parameter: {key} = {keyword_parameters[key]!r}.'
            )
        
        if key_rule_trigger_type is None:
            continue
        
        if rule_trigger_type is AutoModerationRuleTriggerType.none:
            rule_trigger_type = key_rule_trigger_type
            continue
        
        if rule_trigger_type is not key_rule_trigger_type:
            raise TypeError(
                f'Multiple rule trigger types received / detected: {rule_trigger_type!r}; {key_rule_trigger_type!r}; '
                f'keyword parameters = {keyword_parameters!r}.'
            )
    
    return rule_trigger_type



__all__ = ()

from ..audit_log_entry_detail_conversion import AuditLogEntryDetailConversion, AuditLogEntryDetailConversionGroup


# ---- days ----

DAYS_CONVERSION = AuditLogEntryDetailConversion(
    'delete_member_days',
    'days',
)


@DAYS_CONVERSION.set_value_deserializer
def days_value_deserializer(value):
    if value is None:
        value = 0
    return value


@DAYS_CONVERSION.set_value_validator
def days_value_validator(value):
    if not isinstance(value, int):
        raise TypeError(f'`days` can be `int`, got {value.__class__.__name__}; {value!r}.')
    
    if value < 0:
        raise ValueError(f'`days` cannot be negative, got {value!r}.')
    
    return value


# ---- users_removed ----

USERS_REMOVED_CONVERSION = AuditLogEntryDetailConversion(
    'members_removed',
    'users_removed',
)


@USERS_REMOVED_CONVERSION.set_value_deserializer
def users_removed_value_deserializer(value):
    if value is None:
        value = 0
    return value


@USERS_REMOVED_CONVERSION.set_value_validator
def users_removed_value_validator(value):
    if not isinstance(value, int):
        raise TypeError(f'`users_removed` can be `int`, got {value.__class__.__name__}; {value!r}.')
    
    if value < 0:
        raise ValueError(f'`users_removed` cannot be negative, got {value!r}.')
    
    return value


# Construct

GUILD_CONVERSIONS = AuditLogEntryDetailConversionGroup(
    DAYS_CONVERSION,
    USERS_REMOVED_CONVERSION,
)



__all__ = ('GuildActivityOverview',)

from ...bases import DiscordEntity, IconSlot
from ...http.urls import (
    build_guild_badge_icon_url, build_guild_badge_icon_url_as, build_guild_discovery_splash_url,
    build_guild_discovery_splash_url_as, build_guild_icon_url, build_guild_icon_url_as
)
from ...precreate_helpers import process_precreate_parameters_and_raise_extra
from ...scheduled_event import PrivacyLevel

from ..guild_activity_overview_activity import GuildActivityOverviewActivity, GuildActivityOverviewActivityLevel

from .fields import (
    parse_activities, parse_activity_application_ids, parse_approximate_online_count, parse_approximate_user_count,
    parse_badge, parse_badge_color_primary, parse_badge_color_secondary, parse_badge_tag, parse_banner_color,
    parse_boost_count, parse_boost_level, parse_description, parse_features, parse_id, parse_name, parse_privacy_level,
    parse_tags, put_activities, put_activity_application_ids, put_approximate_online_count, put_approximate_user_count,
    put_badge, put_badge_color_primary, put_badge_color_secondary, put_badge_tag, put_banner_color, put_boost_count,
    put_boost_level, put_description, put_features, put_id, put_name, put_privacy_level, put_tags, validate_activities,
    validate_activity_application_ids, validate_approximate_online_count, validate_approximate_user_count,
    validate_badge, validate_badge_color_primary, validate_badge_color_secondary, validate_badge_tag,
    validate_banner_color, validate_boost_count, validate_boost_level, validate_description, validate_features,
    validate_id, validate_name, validate_privacy_level, validate_tags, 
)

GUILD_ACTIVITY_OVERVIEW_BADGE_ICON = IconSlot('badge_icon', 'badge_hash')
GUILD_ACTIVITY_OVERVIEW_DISCOVERY_SPLASH = IconSlot('discovery_splash', 'custom_banner_hash')
GUILD_ACTIVITY_OVERVIEW_ICON = IconSlot('icon', 'icon')


PRECREATE_FIELDS = {
    'activities' : ('activities', validate_activities),
    'activity_application_ids' : ('activity_application_ids', validate_activity_application_ids),
    'approximate_online_count' : ('approximate_online_count', validate_approximate_online_count),
    'approximate_user_count' : ('approximate_user_count', validate_approximate_user_count),
    'badge' : ('badge', validate_badge),
    'badge_color_primary' : ('badge_color_primary', validate_badge_color_primary),
    'badge_color_secondary' : ('badge_color_secondary', validate_badge_color_secondary),
    'badge_icon': ('badge_icon', GUILD_ACTIVITY_OVERVIEW_BADGE_ICON.validate_icon),
    'badge_tag' : ('badge_tag', validate_badge_tag),
    'banner_color': ('banner_color', validate_banner_color),
    'boost_count' : ('boost_count', validate_boost_count),
    'boost_level' : ('boost_level', validate_boost_level),
    'description' : ('description', validate_description),
    'discovery_splash': ('discovery_splash', GUILD_ACTIVITY_OVERVIEW_DISCOVERY_SPLASH.validate_icon),
    'features' : ('features', validate_features),
    'icon' : ('icon', GUILD_ACTIVITY_OVERVIEW_ICON.validate_icon),
    'name' : ('name', validate_name),
    'privacy_level': ('privacy_level', validate_privacy_level),
    'tags' : ('tags', validate_tags),
}



def _application_id_and_activity_sort_key_getter(item):
    """
    Sort key getter used when sorting ``GuildActivityOverviewActivity``-s in display order.
    
    Parameters
    ----------
    item : ``((int, int, int), (int, GuildActivityOverviewActivity))``
        Item to get key of.
    
    Returns
    -------
    key : `(int, int, int)`
    """
    return item[0]


class GuildActivityOverview(DiscordEntity):
    """
    Represents a guild's activity's overview.
    
    Attributes
    ----------
    activities : `None | dict<int, GuildActivityOverviewActivity>`
        An application identifier to an activity overview relation.
        Its application identifiers are a subset of the ones mentioned in ``.activity_application_ids``.
    
    activity_application_ids : `None | tuple<int>`
        Application identifiers of activities to be shown.
    
    approximate_online_count : `int`
        Approximate amount of online users at the represented guild.
    
    approximate_user_count : `int`
        Approximate amount of users at the represented guild.
    
    badge : `int`
        Unknown type or flag field.
        
        `0` means no badge. Likely values: `0`, `2`, `6`, `7`.
    
    badge_color_primary : `None | Color`
        Auto generated primary banner color.
    
    badge_color_secondary : `None | Color`
        Auto generated secondary banner color.
    
    badge_icon_hash : `int`
        The guild's badge icon's hash in `uint128`.
    
    badge_icon_type : ``IconType``
        The guild's badge icon's type.
    
    badge_tag : `str`
        The guild's badge tag.
    
    banner_color : `None | Color`
        Banner color.
    
    boost_count : `int`
        The amount of boosts the represented guild has.
    
    boost_level : `int`
        The boost level of the represented guild.
    
    description : `None | str`
        Description of the represented guild.
    
    discovery_splash_hash : `int`
        The guild's discovery splash's hash in `uint128`.
    
    discovery_splash_type : ``IconType``
        The guild's discovery splash's type.
    
    features :  ``None | tuple<GuildFeature>``
        The represented guild's public features.
    
    icon_hash : `int`
        The guild's icon's hash in `uint128`.
    
    icon_type : ``IconType``
        The guild's icon's type.
    
    id : `int`
        The represented guild's identifier.
    
    name : `str`
        Name of the represented guild.
    
    privacy_level : ``PrivacyLevel``
        For who is the guild overview visible for and other related information.
    
    tags : ``None | tuple<GuildActivityOverviewTag>``
        Additional tags assigned to the guild.
    """
    __slots__ = (
        'activities', 'activity_application_ids', 'approximate_online_count', 'approximate_user_count',
        'badge', 'badge_color_primary', 'badge_color_secondary', 'badge_tag', 'banner_color', 'boost_count',
        'boost_level', 'description', 'features', 'name', 'privacy_level', 'tags'
    )
    
    discovery_splash = GUILD_ACTIVITY_OVERVIEW_DISCOVERY_SPLASH
    icon = GUILD_ACTIVITY_OVERVIEW_ICON
    badge_icon = GUILD_ACTIVITY_OVERVIEW_BADGE_ICON
    
    def __new__(
        cls,
        *,
        activity_application_ids = ...,
        banner_color = ...,
        description = ...,
        discovery_splash = ...,
        icon = ...,
        name = ...,
        privacy_level = ...,
        tags = ...,
    ):
        """
        Creates a new guild overview.
        
        Parameters
        ----------
        activity_application_ids : ``None | iterable<Application> | iterable<int>``, Optional (Keyword only)
            Application identifiers of activities to be shown.
        
        banner_color : `None | Color | int`, Optional (Keyword only)
            Banner color.
        
        description : `None | str`, Optional (Keyword only)
            Description of the represented guild.
        
        discovery_splash : ``None | str | bytes-like | Icon``, Optional (Keyword only)
            The represented guild's discovery splash.
        
        icon : ``None | str | bytes-like | Icon``, Optional (Keyword only)
            The represented guild's icon.
        
        name : `str`, Optional (Keyword only)
            Name of the represented guild.
        
        privacy_level : ``None | int | PrivacyLevel``, Optional (Keyword only)
            For who is the guild overview visible for and other related information.
        
        tags : ``None | iterable<GuildActivityOverviewTag>``, Optional (Keyword only)
            Additional tags assigned to the guild.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value in incorrect.
        """
        # activity_application_ids
        if activity_application_ids is ...:
            activity_application_ids = None
        else:
            activity_application_ids = validate_activity_application_ids(activity_application_ids)
        
        # banner_color
        if banner_color is ...:
            banner_color = None
        else:
            banner_color = validate_banner_color(banner_color)
        
        # description
        if description is ...:
            description = None
        else:
            description = validate_description(description)
        
        # discovery_splash
        if discovery_splash is ...:
            discovery_splash = None
        else:
            discovery_splash = cls.discovery_splash.validate_icon(discovery_splash, allow_data = True)
        
        # icon
        if icon is ...:
            icon = None
        else:
            icon = cls.icon.validate_icon(icon, allow_data = True)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # privacy_level
        if privacy_level is ...:
            privacy_level = PrivacyLevel.guild_only
        else:
            privacy_level = validate_privacy_level(privacy_level)
        
        # tags
        if tags is ...:
            tags = None
        else:
            tags = validate_tags(tags)
        
        # Construct
        self = object.__new__(cls)
        self.activities = None
        self.activity_application_ids = activity_application_ids
        self.approximate_online_count = 0
        self.approximate_user_count = 0
        self.badge = 0
        self.badge_color_primary = None
        self.badge_color_secondary = None
        self.badge_icon = None
        self.badge_tag = ''
        self.banner_color = banner_color
        self.boost_count = 0
        self.boost_level = 0
        self.description = description
        self.discovery_splash = discovery_splash
        self.features = None
        self.icon = icon
        self.id = 0
        self.name = name
        self.privacy_level = privacy_level
        self.tags = tags
        return self
    
    
    def __repr__(self):
        """Returns repr(self)."""
        repr_parts = ['<', type(self).__name__]
        
        field_added = False
        
        # activities
        activities = self.activities
        if (activities is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' activities = ')
            repr_parts.append(repr(self.activities))
        
        activity_application_ids = self.activity_application_ids
        if (activity_application_ids is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            # activity_application_ids
            repr_parts.append(' activity_application_ids = ')
            repr_parts.append(repr(self.activity_application_ids))
        
        # approximate_online_count & approximate_user_count
        approximate_online_count = self.approximate_online_count
        approximate_user_count = self.approximate_user_count
        if approximate_online_count or approximate_user_count:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            # approximate_online_count
            repr_parts.append(' approximate_online_count = ')
            repr_parts.append(repr(self.approximate_online_count))
            
            # approximate_user_count
            repr_parts.append(', approximate_user_count = ')
            repr_parts.append(repr(self.approximate_user_count))
        
        # badge
        badge = self.badge
        if badge:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            # badge
            repr_parts.append(' badge = ')
            repr_parts.append(repr(badge))
        
        # badge_color_primary
        badge_color_primary = self.badge_color_primary
        if (badge_color_primary is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            # badge_color_primary
            repr_parts.append(' badge_color_primary = ')
            repr_parts.append(repr(badge_color_primary))
        
        # badge_color_secondary
        badge_color_secondary = self.badge_color_secondary
        if (badge_color_secondary is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            # badge_color_secondary
            repr_parts.append(' badge_color_secondary = ')
            repr_parts.append(repr(badge_color_secondary))
        
        # badge_icon
        badge_icon = self.badge_icon
        if badge_icon:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' badge_icon = ')
            repr_parts.append(repr(badge_icon))
        
        # badge_tag
        badge_tag = self.badge_tag
        if badge_tag:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' badge_tag = ')
            repr_parts.append(repr(badge_tag))
        
        # banner_color
        banner_color = self.banner_color
        if (banner_color is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' banner_color = ')
            repr_parts.append(repr(banner_color))
        
        # boost_count
        boost_count = self.boost_count
        if boost_count:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' boost_count = ')
            repr_parts.append(repr(boost_count))
        
        # boost_level
        boost_level = self.boost_level
        if boost_level:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' boost_level = ')
            repr_parts.append(repr(boost_level))
        
        # description
        description = self.description
        if (description is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' description = ')
            repr_parts.append(repr(description))
        
        # discovery_splash
        discovery_splash = self.discovery_splash
        if discovery_splash:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' discovery_splash = ')
            repr_parts.append(repr(discovery_splash))
        
        # features
        features = self.features
        if (features is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' features = ')
            repr_parts.append(repr(features))
        
        # icon
        icon = self.icon
        if icon:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' icon = ')
            repr_parts.append(repr(icon))
        
        # id
        guild_id = self.id
        if guild_id:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' id = ')
            repr_parts.append(repr(guild_id))
        
        # name
        name = self.name
        if name:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' name = ')
            repr_parts.append(repr(name))
        
        # privacy_level
        privacy_level = self.privacy_level
        if privacy_level:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' privacy_level = ')
            repr_parts.append(privacy_level.name)
            repr_parts.append(' ~ ')
            repr_parts.append(repr(privacy_level.value))
        
        # tags
        tags = self.tags
        if tags:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' tags = ')
            repr_parts.append(repr(tags))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns hash(self)."""
        hash_value = 0
        
        # activities
        activities = self.activities
        if (activities is not None):
            hash_value ^= len(activities) << 8
            for application_id, activity in activities.items():
                hash_value ^= application_id & hash(activity)
        
        # activity_application_ids
        activity_application_ids = self.activity_application_ids
        if (activity_application_ids is not None):
            hash_value ^= len(activity_application_ids) << 9
            for application_id in activity_application_ids:
                hash_value ^= application_id
        # approximate_online_count
        approximate_online_count = self.approximate_online_count
        if approximate_online_count:
            hash_value ^= 1 << 20
            hash_value ^= approximate_online_count
        
        # approximate_user_count
        approximate_user_count = self.approximate_user_count
        if approximate_user_count:
            hash_value ^= 1 << 21
            hash_value ^= approximate_user_count
        
        # badge
        badge = self.badge
        if badge:
            hash_value ^= 1 << 21
            hash_value ^= badge
        
        # badge_color_primary
        badge_color_primary = self.badge_color_primary
        if (badge_color_primary is not None):
            hash_value ^= 1 << 10
            hash_value ^= badge_color_primary
        
        # badge_color_secondary
        badge_color_secondary = self.badge_color_secondary
        if (badge_color_secondary is not None):
            hash_value ^= 1 << 23
            hash_value ^= badge_color_secondary
        
        # badge_icon
        badge_icon = self.badge_icon
        if badge_icon:
            hash_value ^= 1 << 18
            hash_value ^= hash(badge_icon)
        
        # badge_tag
        badge_tag = self.badge_tag
        if (badge_tag is not None):
            hash_value ^= 1 << 19
            hash_value ^= hash(badge_tag)
        
        # banner_color
        banner_color = self.banner_color
        if (banner_color is not None):
            hash_value ^= 1 << 24
            hash_value ^= banner_color
        
        # boost_count
        boost_count = self.boost_count
        if boost_count:
            hash_value ^= 1 << 11
            hash_value ^= boost_count
        
        # boost_level
        boost_level = self.boost_level
        if boost_level:
            hash_value ^= 1 << 12
            hash_value ^= boost_level
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= 1 << 13
            hash_value ^= hash(description)
        
        # discovery_splash
        discovery_splash = self.discovery_splash
        if discovery_splash:
            hash_value ^= 1 << 22
            hash_value ^= hash(discovery_splash)
        
        # features
        features = self.features
        if (features is not None):
            hash_value ^= len(features) << 14
            for feature in features:
                hash_value ^= hash(feature)
        
        # icon
        icon = self.icon
        if icon:
            hash_value ^= 1 << 15
            hash_value ^= hash(icon)
        
        # id -> skip
        
        # name
        name = self.name
        if name:
            hash_value ^= 1 << 16
            hash_value ^= hash(name)
        
        # privacy_level
        hash_value ^= hash(self.privacy_level)
        
        # tags
        tags = self.tags
        if (tags is not None):
            hash_value ^= len(tags) << 17
            for tag in tags:
                hash_value ^= hash(tag)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns self != other."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether self equals to other. Other must be the same type as wel.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        is_equal : `bool`
        """
        # id
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            if self_id != other_id:
                return False
            
            # activities
            if self.activities != other.activities:
                return False
            
            # approximate_online_count
            if self.approximate_online_count != other.approximate_online_count:
                return False
            
            # approximate_user_count
            if self.approximate_user_count != other.approximate_user_count:
                return False
            
            # badge
            if self.badge != other.badge:
                return False
            
            # badge_color_primary
            if self.badge_color_primary != other.badge_color_primary:
                return False
            
            # badge_color_secondary
            if self.badge_color_secondary != other.badge_color_secondary:
                return False
            
            # badge_icon
            if self.badge_icon != other.badge_icon:
                return False
            
            # badge_tag
            if self.badge_tag != other.badge_tag:
                return False
            
            # boost_count
            if self.boost_count != other.boost_count:
                return False
            
            # boost_level
            if self.boost_level != other.boost_level:
                return False
            
            # features
            if self.features != other.features:
                return False
        
        # activity_application_ids
        if self.activity_application_ids != other.activity_application_ids:
            return False
        
        # banner_color
        if self.banner_color != other.banner_color:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # discovery_splash
        if self.discovery_splash != other.discovery_splash:
            return False
        
        # icon
        if self.icon != other.icon:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # privacy_level
        if self.privacy_level is not other.privacy_level:
            return False
        
        # tags
        if self.tags != other.tags:
            return False
        
        return True
        
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new information.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Channel data receive from Discord.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.activities = parse_activities(data)
        self.activity_application_ids = parse_activity_application_ids(data)
        self.approximate_online_count = parse_approximate_online_count(data)
        self.approximate_user_count = parse_approximate_user_count(data)
        self.badge = parse_badge(data)
        self.badge_color_primary = parse_badge_color_primary(data)
        self.badge_color_secondary = parse_badge_color_secondary(data)
        self._set_badge_icon(data)
        self.badge_tag = parse_badge_tag(data)
        self.banner_color = parse_banner_color(data)
        self.boost_count = parse_boost_count(data)
        self.boost_level = parse_boost_level(data)
        self.description = parse_description(data)
        self._set_discovery_splash(data)
        self.features = parse_features(data)
        self._set_icon(data)
        self.id = parse_id(data)
        self.name = parse_name(data)
        self.privacy_level = parse_privacy_level(data)
        self.tags = parse_tags(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Serializes the guild overview.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal field values should be included.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_activity_application_ids(self.activity_application_ids, data, defaults)
        put_banner_color(self.banner_color, data, defaults)
        put_description(self.description, data, defaults)
        type(self).discovery_splash.put_into(self.discovery_splash, data, defaults, as_data = not include_internals)
        type(self).icon.put_into(self.icon, data, defaults, as_data = not include_internals)
        put_name(self.name, data, defaults)
        put_privacy_level(self.privacy_level, data, defaults)
        put_tags(self.tags, data, defaults)
        
        if include_internals:
            put_activities(self.activities, data, defaults)
            put_approximate_online_count(self.approximate_online_count, data, defaults)
            put_approximate_user_count(self.approximate_user_count, data, defaults)
            put_badge(self.badge, data, defaults)
            put_badge_color_primary(self.badge_color_primary, data, defaults)
            put_badge_color_secondary(self.badge_color_secondary, data, defaults)
            type(self).badge_icon.put_into(self.badge_icon, data, defaults, as_data = not include_internals)
            put_badge_tag(self.badge_tag, data, defaults)
            put_boost_count(self.boost_count, data, defaults)
            put_boost_level(self.boost_level, data, defaults)
            put_features(self.features, data, defaults)
            put_id(self.id, data, defaults)
        
        return data
    
    
    @classmethod
    def _create_empty(cls, guild_id):
        """
        Creates an empty guild activity overview with the given fields.
        
        Parameters
        ----------
        guild_id : `int`
            Guild identifier to create the guild activity overview with.
        
        Returns
        -------
        self : `instance<cls>`
        """
        # Construct
        self = object.__new__(cls)
        self.activities = None
        self.activity_application_ids = None
        self.approximate_online_count = 0
        self.approximate_user_count = 0
        self.badge = 0
        self.badge_color_primary = None
        self.badge_color_secondary = None
        self.badge_icon = None
        self.badge_tag = ''
        self.banner_color = None
        self.boost_count = 0
        self.boost_level = 0
        self.description = None
        self.discovery_splash = None
        self.features = None
        self.icon = None
        self.id = guild_id
        self.name = ''
        self.privacy_level = PrivacyLevel.guild_only
        self.tags = None
        return self
    
    
    @classmethod
    def precreate(cls, guild_id, **keyword_parameters):
        """
        Creates a guild activity overview instance.
        Since these objects are not cached, the only advantage of using ``.precreate`` is that it allows setting
        ``.id`` and various other fields that cannot be used for templating.
        
        Parameters
        ----------
        guild_id : `int`
            The represented guild's identifier.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        activities : `None | dict<int, GuildActivityOverviewActivity>`, Optional (Keyword only)
            An application identifier to an activity overview relation.
        
        activity_application_ids : ``None | iterable<Application> | iterable<int>``, Optional (Keyword only)
            Application identifiers of activities to be shown.
        
        approximate_online_count : `None | int`, Optional (Keyword only)
            Approximate amount of online users at the represented guild.
        
        approximate_user_count : `None | int`, Optional (Keyword only)
            Approximate amount of users at the represented guild.
        
        badge : `None | int`, Optional (Keyword only)
            Unknown type or flag field.
        
        badge_color_primary : `None | int | Color`, Optional (Keyword only)
            Auto generated primary banner color.
        
        badge_color_secondary : `None | int | Color`, Optional (Keyword only)
            Auto generated secondary banner color.
        
        badge_icon : ``None | str | Icon``, Optional (Keyword only)
            The guild's badge icon.
        
        badge_tag : `None | str`, Optional (Keyword only)
            The guild's badge tag.
        
        banner_color : `None | int | Color`, Optional (Keyword only)
            Banner color.
        
        boost_count : `None | int`, Optional (Keyword only)
            The amount of boosts the represented guild has.
        
        boost_level : `None | int`, Optional (Keyword only)
            The boost level of the represented guild.
        
        description : `None | str`, Optional (Keyword only)
            Description of the represented guild.
        
        discovery_splash : ``None | str | Icon``, Optional (Keyword only)
            The represented guild's discovery splash.
        
        features : `None | iterable<GuildFeature> | iterable<str>``, Optional (Keyword only)
            The represented guild's features.
        
        icon : ``None | str | Icon``, Optional (Keyword only)
            The guild's icon.
        
        name : `str`, Optional (Keyword only)
            The represented guild's name.
        
        privacy_level : ``None | int | PrivacyLevel``, Optional (Keyword only)
            For who is the guild overview visible for and other related information.
        
        tags : ``None | iterable<GuildActivityOverviewTag>``, Optional (Keyword only)
            Additional tags assigned to the guild.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - if a parameter's value is incorrect.
        """
        guild_id = validate_id(guild_id)
        
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        self = cls._create_empty(guild_id)
        
        if (processed is not None):
            for name, value in processed:
                setattr(self, name, value)
        
        return self
    
    
    def copy(self):
        """
        Copies the guild overview.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.activities = None
        
        activity_application_ids = self.activity_application_ids
        if (activity_application_ids is not None):
            activity_application_ids = (*activity_application_ids,)
        new.activity_application_ids = activity_application_ids
        
        new.approximate_online_count = 0
        new.approximate_user_count = 0
        new.badge = 0
        new.badge_color_primary = None
        new.badge_color_secondary = None
        new.badge_icon = None
        new.badge_tag = ''
        new.banner_color = self.banner_color
        new.boost_count = 0
        new.boost_level = 0
        new.description = self.description
        new.discovery_splash = self.discovery_splash
        new.features = None
        new.icon = self.icon
        new.id = 0
        new.name = self.name
        new.privacy_level = self.privacy_level
        
        tags = self.tags
        if (tags is not None):
            tags = (*(tag.copy() for tag in tags),)
        new.tags = tags
        
        return new
    
    
    def copy_with(
        self,
        *,
        activity_application_ids = ...,
        banner_color = ...,
        description = ...,
        discovery_splash = ...,
        icon = ...,
        name = ...,
        tags = ...,
        privacy_level = ...,
    ):
        """
        Copies the guild overview with the given fields.
        
        Parameters
        ----------
        activity_application_ids : ``None | iterable<Application> | iterable<int>``, Optional (Keyword only)
            Application identifiers of activities to be shown.
        
        banner_color : `None | Color | int`, Optional (Keyword only)
            Banner color.
        
        description : `None | str`, Optional (Keyword only)
            Description of the represented guild.
        
        discovery_splash : ``None | str | bytes-like | Icon``, Optional (Keyword only)
            The represented guild's discovery splash.
        
        icon : ``None | str | bytes-like | Icon``, Optional (Keyword only)
            The represented guild's icon.
        
        name : `str`, Optional (Keyword only)
            The name of the represented guild.
        
        tags : ``None | iterable<GuildActivityOverviewTag>``, Optional (Keyword only)
            Additional tags assigned to the guild.
        
        privacy_level : ``None | int | PrivacyLevel``, Optional (Keyword only)
            For who is the guild overview visible for and other related information.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value in incorrect.
        """
        # activity_application_ids
        if activity_application_ids is ...:
            activity_application_ids = self.activity_application_ids
            if (activity_application_ids is not None):
                activity_application_ids = (*activity_application_ids,)
        else:
            activity_application_ids = validate_activity_application_ids(activity_application_ids)
        
        # banner_color
        if banner_color is ...:
            banner_color = self.banner_color
        else:
            banner_color = validate_banner_color(banner_color)
        
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # discovery_splash
        if discovery_splash is ...:
            discovery_splash = self.discovery_splash
        else:
            discovery_splash = type(self).discovery_splash.validate_icon(discovery_splash, allow_data = True)
        
        # icon
        if icon is ...:
            icon = self.icon
        else:
            icon = type(self).icon.validate_icon(icon, allow_data = True)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # tags
        if tags is ...:
            tags = self.tags
            if (tags is not None):
                tags = (*(tag.copy() for tag in tags),)
        else:
            tags = validate_tags(tags)
        
        # privacy_level
        if privacy_level is ...:
            privacy_level = self.privacy_level
        else:
            privacy_level = validate_privacy_level(privacy_level)
        
        # Construct
        new = object.__new__(type(self))
        new.activities = None
        new.activity_application_ids = activity_application_ids
        new.approximate_online_count = 0
        new.approximate_user_count = 0
        new.badge = 0
        new.badge_color_primary = None
        new.badge_color_secondary = None
        new.badge_icon = None
        new.badge_tag = ''
        new.banner_color = banner_color
        new.boost_count = 0
        new.boost_level = 0
        new.description = description
        new.discovery_splash = discovery_splash
        new.features = None
        new.icon = icon
        new.id = 0
        new.name = name
        new.privacy_level = privacy_level
        new.tags = tags
        return new
    
    
    def iter_application_ids_and_activities_ordered(self):
        """
        Iterates over the represented application identifiers and the overview activities in display order.
        
        This method is an iterable generator.
        
        Yields
        ------
        application_id_and_activity : ``(int, GuildActivityOverviewActivity)``
        """
        activity_application_ids = self.activity_application_ids
        if activity_application_ids is None:
            return
        
        activities = self.activities
        if (activities is None):
            for application_id in activity_application_ids:
                yield application_id, GuildActivityOverviewActivity._create_empty()
            return
        
        sort_key_output_pairs = []
        
        for index, application_id in enumerate(activity_application_ids):
            try:
                activity = activities[application_id]
            except KeyError:
                activity = GuildActivityOverviewActivity._create_empty()
            
            activity_level = activity.level
            if activity_level is GuildActivityOverviewActivityLevel.recently_popular:
                priority = 0
            
            elif activity_level is GuildActivityOverviewActivityLevel.any_previous:
                priority = 1
            
            else:
                priority = 2
            
            sort_key_output_pairs.append(((priority, -activity.score, index), (application_id, activity)))
        
        sort_key_output_pairs.sort(key = _application_id_and_activity_sort_key_getter)
        for item in sort_key_output_pairs:
            yield item[1]
    
    
    def iter_activity_application_ids(self):
        """
        Iterates over the activity application identifiers.
        
        This method is an iterable generator.
        
        Yields
        ------
        activity_application_id : `int`
        """
        activity_application_ids = self.activity_application_ids
        if (activity_application_ids is not None):
            yield from activity_application_ids
    
    
    def iter_features(self):
        """
        Iterates over the represented guild's features.
        
        This method is an iterable generator.
        
        Yields
        ------
        feature : ``GuildFeature``
        """
        features = self.features
        if (features is not None):
            yield from features
    
    
    def iter_tags(self):
        """
        Iterates over the guild activity overview's tags.
        
        This method is an iterable generator.
        
        Yields
        ------
        tag : ``GuildActivityOverviewTag``
        """
        tags = self.tags
        if (tags is not None):
            yield from tags
    
    
    @property
    def badge_icon_url(self):
        """
        Returns the guild badge's icon's url. If the guild badge has no icon, then returns `None`.
        
        Returns
        -------
        url : `None | str`
        """
        return build_guild_badge_icon_url(self.id, self.badge_icon_type, self.badge_icon_hash)
    
    
    def badge_icon_url_as(self, ext = None, size = None):
        """
        Returns the guild badge's icon's url. If the guild badge has no icon, then returns `None`.
        
        Parameters
        ----------
        ext : `None | str` = `None`, Optional
            The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
            If the guild has animated badge icon, it can be `'gif'` as well.
        
        size : `None | int` = `None`, Optional
            The preferred minimal size of the image's url.
        
        Returns
        -------
        url : `None | str`
        """
        return build_guild_badge_icon_url_as(self.id, self.badge_icon_type, self.badge_icon_hash, ext, size)
    
    
    @property
    def discovery_splash_url(self):
        """
        Returns the guild's discovery splash's url. If the guild has no discovery_splash, then returns `None`.
        
        Returns
        -------
        url : `None | str`
        """
        return build_guild_discovery_splash_url(self.id, self.discovery_splash_type, self.discovery_splash_hash)
    
    
    def discovery_splash_url_as(self, ext = None, size = None):
        """
        Returns the guild's discovery splash's url. If the guild has no discovery splash, then returns `None`.
        
        Parameters
        ----------
        ext : `None | str` = `None`, Optional
            The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
            If the guild has animated discovery splash, it can be `'gif'` as well.
        
        size : `None | int` = `None`, Optional
            The preferred minimal size of the image's url.
        
        Returns
        -------
        url : `None | str`
        """
        return build_guild_discovery_splash_url_as(self.id, self.discovery_splash_type, self.discovery_splash_hash, ext, size)
    
    
    @property
    def icon_url(self):
        """
        Returns the guild's icon's url. If the guild has no icon, then returns `None`.
        
        Returns
        -------
        url : `None | str`
        """
        return build_guild_icon_url(self.id, self.icon_type, self.icon_hash)
    
    
    def icon_url_as(self, ext = None, size = None):
        """
        Returns the guild's icon's url. If the guild has no icon, then returns `None`.
        
        Parameters
        ----------
        ext : `None | str` = `None`, Optional
            The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
            If the guild has animated icon, it can `'gif'` as well.
        
        size : `None | int` = `None`, Optional
            The preferred minimal size of the image's url.
        
        Returns
        -------
        url : `None | str`
        """
        return build_guild_icon_url_as(self.id, self.icon_type, self.icon_hash, ext, size)
